# -*- coding: utf-8 -*-

import numpy as np
import matplotlib.pyplot as plt
import matplotlib
import time
from mpmath import mp 
import mpmath.libmp
mpmath.libmp.BACKEND
import os
import sys
import shutil
import pandas as pd
import ipynb_path


######################### START OF HEADER #########################
######################### START OF HEADER #########################
######################### START OF HEADER #########################
######################### START OF HEADER #########################
######################### START OF HEADER #########################

spi = mp.sqrt(mp.pi)

#>>> Set floating point precision　<<<
#*** mp.dps determines the number of significant figure.
#*** Reducing mp.dps slightly accelerate the computation.
#*** If mp.dps is reduced from the default value of 200, 
#*** increase the value of "small" accordingly.
mp.dps = 200
small = mp.mpf('1e-32') # the value of 0+

#>>> Name of input file <<<
#*** The input file should be located in the "current directory".
#*** To check the name of the current directory, 
#*** just run the code and see the comment in the console.
#*** "id" must be six characters.
id = 'b00100'

#>>> Where to display the results <<<
#*** Choose 'y' to see the results in the console.
#*** Choose 'n' to export the results in a folder XXXXXX,
#*** where XXXXXX is "id" defined below.
display_macro_val_on_console = 'n'

#>>> How to read input parameters <<<
#*** Choose 'y' to enter parameters manually one by one.
#*** Choose 'n' to import the parameters from an input file,
#*** which must be prepared in advance.
manual_enter_parameter = 'n'

#>>> Trancate the solid parts partially <<< 
#*** This works only when "magnify == 'n'"
#*** "H_trnc" is the distance from the fluid-solid interface.
#*** That is, z in the range [-H_tranc, H + H_tranc] will be displayed.
truncate = 'n'; H_trnc = mp.mpf('5e-6')

#>>> Choose the domain of interest <<< 
#*** Choose 'n' to compute the entire domain
#*** Choose 'y' to magnify a specific region
#*** If 'y' is chosen, determine the size of the region 
#*** with the unit of meter, so that 
#*** z in [z_min,z_max] and r in [r_min,r_max] becomes 
#*** the region of interest. 
magnify = 'y'
r_min = mp.mpf('0e-6'); r_max = mp.mpf('40e-6')
z_min = mp.mpf('0e-6'); z_max = mp.mpf('2e-6')

#>>> Whether or not to show non-dimensional fields <<<
#*** Showing non-dimensional fields takes some seconds. 
#*** If not necessary, this may be turned off.
show_nd = 'y'

#>>> Automatic search for a nice value of s_max <<<
#*** Increase "s_max" until the error indicator 
#*** "chk_s" becomes lower than "threshold_1_s_max" iteratively. 
search_s_max = 'n'
threshold_1_s_max = 1e-2 # Decrease when more accurate solution is needed.
threshold_2_s_max = 1e-5 
magnification_factor = mp.mpf('2')  
Nr_ss = 3; Nz_ss = 3
Nm_ss = 8 # Default = 8; Use a larger value when (code w01) occurs
pmax  = 100 # Max number of the iteration

#>>> Automatic search for a nice value of Nm <<<
#*** Increase "Nm" until the error indicator 
#*** "chk_Nm" becomes lower than "threshold_1_Nm" iteratively. 
search_Nm = 'n'
threshold_1_Nm = 1e-2
magnification_factor_Nm = mp.sqrt('2')
Nr_sN = 3; Nz_sN = 3; Nm_sN = 8
pp = -1

#>>> Layout handling <<<
#*** Choose 'constrained' to automatically adjust nice layout.
#*** Either choose 'tight'.
#*** 'constrained' may cause error,
#*** depending on the aspect ratio of the region of interest.
hdl_layout = 'none'
hdl_layout = 'constrained'

#>>> Lock the aspect ratio of the figure <<<'
hdl_aspect = 'auto' 
hdl_aspect = 'equal'

#>>> Contour setting <<<    
#*** The number of contour lines
ct_lv   = 20 
#*** The size of colormap legend
clb_shk = 0.8

#>>> Discretization in s-space <<<
#*** "s" is discretized as s = s_max * [(m-1/(Nm-1))**power]. 
#*** (Note that m=1 -> s=0 and m=Nm -> s=s_max)
#*** The upper end "s_max" is controlled "s_max0" in the input file.
#*** If "power" is 1, "s" is uniformly discretized (not recommended).
#*** Choosing larger "power" means the denser grid points near s = 0, 
#*** and may allow the use of small number of "Nm".
#*** Default is "power = 2".
#*** If smaller Nm is required, try to use "power = 3" or larger.
power = mp.mpf('2') 

# check for versions in the current setup
show_version = 'y'

#>>> Avoiding the occurence of a very large number <<<
#*** When "warning code w01" is not resolved even by 
#*** the use of large s_max0, try to increase "upper_limit". 
#*** This may occur when very large H_1 and/or H_2 is used. 
#*** Basically, there is no need to change.
upper_limit = mp.mpf('11180')

######################### END OF HEADER #########################
######################### END OF HEADER #########################
######################### END OF HEADER #########################
######################### END OF HEADER #########################
######################### END OF HEADER #########################


# Open dialog to read input file (not used)
def get_file(): 
    #Note tkinter does not work on pyscript
    import tkinter as Tkinter, tkinter.filedialog as tkFileDialog
    root = Tkinter.Tk()
    root.after(100, root.focus_force)
    root.after(200,root.withdraw)    
    file_path = tkFileDialog.askopenfilename(parent=root,title='Pick a file')    
    return file_path 

# Get the name of input file "input_file", 
# and set the file name of the exported files "fname"
def get_fname(id):
    global target_url
    path = os.getcwd() #get current directory
    print('current directory = ',path)
    if len(id) != 6:
        sys.exit('id length mush be 6')
    input_file = "00input-"+id+".txt" 
    fname = id+'/OUTPUT-'+id
    if manual_enter_parameter == 'n':
        if not os.path.isdir(id):
                os.makedirs(id)
        pyfilename = ipynb_path.get()
        #chdir(os.path.dirname(__file__))
        dept =  pyfilename#os.path.basename(__file__)
        dest = id+'/'+"script.py"
        shutil.copy(dept,dest)
        dept = input_file
        dest = id+'/'+input_file
        shutil.copy(dept,dest)
    else:
        print('enter fullpath to input file')
    return input_file, fname

# Main script
def main() -> None:
    
    lattice() # Define lattice 
    compute_mac() # Compute the macroscopic quantities on the defined lattice
    if show_nd == 'y':
        plot_result_nd() # Plot non-dimensional quantities 
    plot_result() # Compute and plot dimensional quantities
    finalize() # Compute some important quantities

# "err" is used in "search_s_max".
# If "err > threshold_1_s_max", this indicates "s_max" is too small. 
# If "err < threshold_2_s_max", this indicates "s_max" can be loosen. 
# "err_Nm" is used in "search_Nm".
def finalize() -> None:
    global err, err_Nm
    err = chk_s()
    if search_Nm == 'y':
        if pp > 0:
            err_Nm = chk_Nm() #this can be doable in search_Nm mode
    if search_s_max == 'n' and search_Nm == 'n':
        print('>> [CHECK] err =',"{:.1e}".format(float(chk_s())), \
              '( s_max = ',"{:.1e}".format(float(s_max)),')')
    if float(err) < threshold_2_s_max:
        if search_s_max == 'n' and search_Nm == 'n':
            print('>> [COMMENT] "s_max0" may be reduced')
    if float(err) > threshold_1_s_max:
        if search_s_max == 'n' and search_Nm == 'n':
            print('\033[33m'+'>> [WARNING] (code w01) "s_max0" may be too small'+'\033[0m')
            

# Reading physical and computational parameters
def read_parameters() -> None:
    global kappa, kappa_1, kappa_2, kappa_m1, kappa_m2
    global rho, C_p     
    global Aa, Ab1, Ab2     
    global P, w_0, z_0, Lambda
    global g, T_0, betac   
    global H, H_1, H_2, H_m1, H_m2    
    global nu      
    global Nr, Aspect  
    global Nz, Nz_1, Nz_2    
    global Nm, case, s_max0  
    global save1, threads 
    global D_T, radius  
    global Kd1, Kd2

    if manual_enter_parameter == 'n':
        f = open(input_file, "r")
        param = np.zeros(0) #prepare empty 1d array
        for lines in f:
            #read from input_file line by line 
            param = np.append(param,lines.split()[0])
    else:
        pname = []
        pname = np.append(pname, 'kappa    ')
        pname = np.append(pname, 'kappa_1  ')
        pname = np.append(pname, 'kappa_2  ')
        pname = np.append(pname, 'kappa_m1 ')
        pname = np.append(pname, 'kappa_m2 ')
        pname = np.append(pname, 'rho      ')
        pname = np.append(pname, 'C_p      ')
        pname = np.append(pname, 'Aa       ')
        pname = np.append(pname, 'Ab1      ')
        pname = np.append(pname, 'Ab2      ')
        pname = np.append(pname, 'P        ')
        pname = np.append(pname, 'w_0      ')
        pname = np.append(pname, 'z_0      ')
        pname = np.append(pname, 'Lambda   ')
        pname = np.append(pname, 'g        ')
        pname = np.append(pname, 'T_0      ')
        pname = np.append(pname, 'betac    ')
        pname = np.append(pname, 'H        ')
        pname = np.append(pname, 'H_1      ')
        pname = np.append(pname, 'H_2      ')
        pname = np.append(pname, 'H_m1     ')
        pname = np.append(pname, 'H_m2     ')
        pname = np.append(pname, 'nu       ')
        pname = np.append(pname, 'Nr       ')
        pname = np.append(pname, 'Aspect   ')
        pname = np.append(pname, 'Nz       ')
        pname = np.append(pname, 'Nz_1     ')
        pname = np.append(pname, 'Nz_2     ')
        pname = np.append(pname, 'Nm       ')
        pname = np.append(pname, 'case     ')
        pname = np.append(pname, 's_max0   ')
        pname = np.append(pname, 'save1    ')
        pname = np.append(pname, 'threads  ')
        pname = np.append(pname, 'D_T      ')
        pname = np.append(pname, 'radius   ')
        pname = np.append(pname, 'Kd1      ')
        pname = np.append(pname, 'Kd2      ')
        param = np.zeros(0)
        for pa in range(37):
            print(pname[pa],end='=')
            param = np.append(param,input())

    #define physical parameters
    kappa    = mp.mpf(param[0] )
    kappa_1  = mp.mpf(param[1] )
    kappa_2  = mp.mpf(param[2] )
    kappa_m1 = mp.mpf(param[3] )
    kappa_m2 = mp.mpf(param[4] )
    rho      = mp.mpf(param[5] )
    C_p      = mp.mpf(param[6] )
    Aa       = mp.mpf(param[7] )
    Ab1      = mp.mpf(param[8] )
    Ab2      = mp.mpf(param[9] )
    P        = mp.mpf(param[10])
    w_0      = mp.mpf(param[11])
    z_0      = mp.mpf(param[12])
    Lambda   = mp.mpf(param[13])
    g        = mp.mpf(param[14])
    T_0      = mp.mpf(param[15])
    betac    = mp.mpf(param[16])
    H        = mp.mpf(param[17])
    H_1      = mp.mpf(param[18])
    H_2      = mp.mpf(param[19])
    H_m1     = mp.mpf(param[20])
    H_m2     = mp.mpf(param[21])
    nu       = mp.mpf(param[22])
    Nr       =    int(param[23])
    Aspect   = mp.mpf(param[24])
    Nz       =    int(param[25])
    Nz_1     =    int(param[26])
    Nz_2     =    int(param[27])
    Nm       =    int(param[28])
    case     =    int(param[29])
    s_max0   = mp.mpf(param[30])
    save1    =    int(param[31])
    threads  =    int(param[32])
    D_T      = mp.mpf(param[33])
    radius   = mp.mpf(param[34])
    Kd1      = mp.mpf(param[35])
    Kd2      = mp.mpf(param[36])
    if magnify == 'y':
        print('>> [NOTE] magnification mode')
        print('>> [NOTE] Nz_1 and Nz_2 are forced to be 1')
        print('>> [NOTE] only fluid domain is computed')
        time.sleep(1)
        Nz_1 = 1
        Nz_2 = 1
    if case == 1:#simpson's rule
        if Nm%2 != 1:
            print('>> [NOTE] simpson rule (case = 1) force Nm to odd' )
            print('>> [NOTE] Nm -> Nm + 1' )
            Nm = Nm + 1
        
# Compute the non-dimensional parameters
def initialize() -> None:
    #define non-dimensional parameter
    global z_R, L_0, I_0      
    global delta_Ta, delta_Tb1, delta_Tb2
    global v_0      
    global delta_uc, delta_ud1, delta_ud2
    global Re, p_0  
    global tH, tH_1, tH_2, tH_m1, tH_m2
    global tk_1, tk_2, tk_m1, tk_m2 
    global tz_R, tz_R2, tz_R4, tz_0 
    global tD_T, F_0  
    global s_max
    z_R       = ( mp.pi * w_0 * w_0 ) / Lambda
    L_0       = w_0
    I_0       = mp.mpf('2') * P / ( mp.pi * w_0 * w_0 )
    delta_Ta  = I_0 * L_0 * L_0 / ( kappa * T_0 ) * Aa 
    delta_Tb1 = I_0 * L_0 * L_0 / ( kappa * T_0 ) * Ab1
    delta_Tb2 = I_0 * L_0 * L_0 / ( kappa * T_0 ) * Ab2
    v_0       = nu / w_0
    delta_uc  = ( g * w_0 * w_0 * betac * T_0 / nu ) / v_0
    delta_ud1 = ( Kd1 * T_0 / w_0 ) / v_0
    delta_ud2 = ( Kd2 * T_0 / w_0 ) / v_0    
    Re        = max( mp.fabs(delta_uc), \
                     mp.fabs(delta_ud1),\
                     mp.fabs(delta_ud2) )
    print('>> [CHECK] Re = ',"{:.2e}".format(float(Re)))
    p_0       = rho * v_0 * v_0
    tH        = H    / L_0
    tH_1      = H_1  / L_0
    tH_2      = H_2  / L_0
    tH_m1     = H_m1 / L_0
    tH_m2     = H_m2 / L_0
    tk_1      = kappa_1  / kappa
    tk_2      = kappa_2  / kappa
    tk_m1     = kappa_m1 / kappa
    tk_m2     = kappa_m2 / kappa
    tz_R      = z_R / L_0
    tz_R2     = tz_R  * tz_R
    tz_R4     = tz_R2 * tz_R2
    tz_0      = z_0 / L_0
    tD_T      = D_T / ( v_0 * L_0 / T_0 )
    F_0       = mp.mpf('6') * mp.pi * radius * ( rho * nu ) *v_0
    s_max     = min( ( z_R / H ) * s_max0, \
                     upper_limit / tH_1,   \
                     upper_limit / tH_2 ) 

def lattice() -> None:
    global tr, tz, tz_1, tz_2, s, weight
    global tr_min, tr_max, tz_min, tz_max
    
    if magnify =='y':
        if search_s_max == 'n' and search_Nm == 'n':
            tr_min = r_min/w_0; tr_max = r_max/w_0
            tz_min = z_min/w_0; tz_max = z_max/w_0
            print('    tr_min = ',mp.nstr(tr_min))
            print('    tr_max = ',mp.nstr(tr_max))
            print('    tz_min = ',mp.nstr(tz_min))
            print('    tz_max = ',mp.nstr(tz_max))
        elif magnify_org == 'y':
            tr_min = r_min/w_0; tr_max = r_max/w_0
            tz_min = z_min/w_0; tz_max = z_max/w_0
        else:
            tr_min = mp.mpf('0')/w_0; tr_max = H*Aspect/w_0
            tz_min = mp.mpf('0')/w_0; tz_max = H/w_0
        if tr_min < mp.mpf('0'):
            sys.exit("magnification r range error, only fluid domain is allowd")
        if tr_max > tH * Aspect + small:
            print('tr_max=',mp.nstr(tr_max))
            print('tH*Aspect_max =',mp.nstr(tH*Aspect))
            sys.exit("magnification r range error, only fluid domain is allowd")
        if tz_min < mp.mpf('0'):
            sys.exit("magnification z range error, only fluid domain is allowd")
        if tz_max > tH: 
            sys.exit("magnification z range error, only fluid domain is allowd")
        tr   = mp.matrix(1,Nr) #tr[0], ..., tr[Nr-1]
        tz   = mp.matrix(1,Nz) #tz[0], ..., tz[Nz-1]
        tz_1 = mp.matrix(1,Nz_1) 
        tz_2 = mp.matrix(1,Nz_2) 
        for i in range(1,Nr+1):
            tr[i-1] = (mp.mpf(i)-mp.mpf('1'))/(mp.mpf(Nr)-mp.mpf('1')) \
                      * (tr_max-tr_min) + tr_min + small

        for j in range(1,Nz+1):
            tz[j-1] = (mp.mpf(j)-mp.mpf('1'))/(mp.mpf(Nz)-mp.mpf('1')) \
                      * (tz_max-tz_min) + tz_min

    else:
        tr   = mp.matrix(1,Nr) #tr[0], ..., tr[Nr-1]
        tz   = mp.matrix(1,Nz) #tz[0], ..., tz[Nz-1]
        tz_1 = mp.matrix(1,Nz_1) 
        tz_2 = mp.matrix(1,Nz_2) 
        for i in range(1,Nr+1):#1, ..., Nr
            tr[i-1] = (mp.mpf(i)-mp.mpf('1'))/(mp.mpf(Nr)-mp.mpf('1')) \
                      * tH * Aspect + small
            
        for j in range(1,Nz+1):#1, ..., Nz
            tz[j-1] = (mp.mpf(j)-mp.mpf('1'))/(mp.mpf(Nz)-mp.mpf('1')) * tH
        
        if truncate == 'n':
            for j in range(1,Nz_1+1):#1, ..., Nz_1
                tz_1[j-1] = (mp.mpf(j)-mp.mpf('1'))/(mp.mpf(Nz_1)-mp.mpf('1')) * tH_1 - tH_1
            for j in range(1,Nz_2+1):#1, ..., Nz_2
                tz_2[j-1] = (mp.mpf(j)-mp.mpf('1'))/(mp.mpf(Nz_2)-mp.mpf('1')) * tH_2 + tH
        else:
            tH_trnc = H_trnc / w_0
            for j in range(1,Nz_1+1):#1, ..., Nz_1
                tz_1[j-1] = (mp.mpf(j)-mp.mpf('1'))/(mp.mpf(Nz_1)-mp.mpf('1')) * tH_trnc - tH_trnc
            for j in range(1,Nz_2+1):#1, ..., Nz_2
                tz_2[j-1] = (mp.mpf(j)-mp.mpf('1'))/(mp.mpf(Nz_2)-mp.mpf('1')) * tH_trnc + tH        
    
    #lattice for s
    s      = mp.matrix(1,Nm) #s[0], ..., s[Nm-1]
    weight = mp.matrix(1,Nm)
    for m in range(1,Nm+1)    :#1, ..., Nm
         s[m-1] = (( (mp.mpf(m)-mp.mpf('1'))/(mp.mpf(Nm)-mp.mpf('1')) \
                     )**power) * s_max + small
            
    if case == 1:# use simpson's rule
        for m in range(2,Nm,2)    :#2, 4, ..., Nm-1
            s[m-1] = ( s[m] + s[m-2] ) * mp.mpf('0.5')
                
    # print(mp.nstr(s))
    # sys.exit()
                        
    for m in range(1,Nm+1)    :#1, ..., Nm        
        #for inverse Hankel
        if case == 2:#trapezoidal rule
            if m == 1:
                weight[m-1] = mp.mpf('0.5') * ( s[m]   - s[m-1] )
            elif m == Nm:
                weight[m-1] = mp.mpf('0.5') * ( s[m-1] - s[m-2] )
            else:
                weight[m-1] = mp.mpf('0.5') * ( s[m]   - s[m-2] )
        elif case == 1:#simpson's rule
            if m%2 == 0:#2, 4, ..., Nm-1
                weight[m-1] = mp.mpf('4')/mp.mpf('6') * ( s[m]   - s[m-2] )
            elif m == 1:
                weight[m-1] = mp.mpf('1')/mp.mpf('6') * ( s[m+1] - s[m-1] )
            elif m == Nm:
                weight[m-1] = mp.mpf('1')/mp.mpf('6') * ( s[m-1] - s[m-3] )
            else: 
                weight[m-1] = mp.mpf('1')/mp.mpf('6') * ( s[m+1] - s[m-3] )
            

                
    
####
def plot_result_nd() -> None:
    
    #whole temperature field (non-dimensional, tau^{(x)})
    #this is skipped in magnification mode
    if search_s_max == 'n' and search_Nm == 'n':
        print('>> [PLOT] non-dimensional temperature')  
    if magnify != 'y':
        for j in range(1,Nz_1+1): 
            tz_all[j-1] = tz_1[j-1]
        for j in range(1,Nz+1): 
            tz_all[j-1+Nz_1] = tz[j-1]
        for j in range(1,Nz_2+1): 
            tz_all[j-1+Nz+Nz_1] = tz_2[j-1]
        x, y = np.meshgrid(tr, tz_all) 
        for i in range(1,Nr+1):
            for j in range(1,Nz_1+1):
                arg = j-1
                z_tau_a_all [arg,i-1] = tau_a_1 [i-1,j-1]
                z_tau_b1_all[arg,i-1] = tau_b1_1[i-1,j-1]
                z_tau_b2_all[arg,i-1] = tau_b2_1[i-1,j-1]
            for j in range(1,Nz+1):
                arg = j-1 + Nz_1
                z_tau_a_all [arg,i-1] = tau_a [i-1,j-1]
                z_tau_b1_all[arg,i-1] = tau_b1[i-1,j-1]
                z_tau_b2_all[arg,i-1] = tau_b2[i-1,j-1]
            for j in range(1,Nz_2+1):
                arg = j-1 + Nz + Nz_1
                z_tau_a_all [arg,i-1] = tau_a_2 [i-1,j-1]
                z_tau_b1_all[arg,i-1] = tau_b1_2[i-1,j-1]
                z_tau_b2_all[arg,i-1] = tau_b2_2[i-1,j-1]
        fig, axs = plt.subplots(1, 3, layout=hdl_layout, figsize=(12,6), dpi=300)            
        fig.suptitle(\
"non-dimensional temperature field (w/o delta_Tx) \
for fluid, solid 1, solid 2 domains \
in (tilde{r},tilde{z})-space"\
)
        panel_tau_a  = axs[0].contour (x,y,z_tau_a_all ,linewidths=0.5,colors='k',levels=20)
        panel_tau_b1 = axs[1].contour (x,y,z_tau_b1_all,linewidths=0.5,colors='k',levels=20)
        panel_tau_b2 = axs[2].contour (x,y,z_tau_b2_all,linewidths=0.5,colors='k',levels=20)
        panel_tau_a  = axs[0].contourf(x,y,z_tau_a_all ,cmap='RdBu_r',levels=20)
        panel_tau_b1 = axs[1].contourf(x,y,z_tau_b1_all,cmap='RdBu_r',levels=20)
        panel_tau_b2 = axs[2].contourf(x,y,z_tau_b2_all,cmap='RdBu_r',levels=20)
        for ab in range(3):
            axs[ab].axhline(0, color='black',ls='--', linewidth=1.0)
            axs[ab].axhline(tH,color='black',ls='--', linewidth=1.0)
            axs[ab].set_aspect(hdl_aspect)
        fig.colorbar(panel_tau_a , ax=axs[0], shrink = clb_shk)
        fig.colorbar(panel_tau_b1, ax=axs[1], shrink = clb_shk)
        fig.colorbar(panel_tau_b2, ax=axs[2], shrink = clb_shk)
        axs[0].set_title('tau_a' )
        axs[1].set_title('tau_b1')
        axs[2].set_title('tau_b2')
        if display_macro_val_on_console == 'n':
            plt.savefig(fname+'_nondimensional-temperature.png')
        plt.show()

    
    #fluid domain only
    if search_s_max == 'n' and search_Nm == 'n':
        print('>> [PLOT] non-dimensional flows')  
    x, y        = np.meshgrid(tr, tz) 
    for i in range(1,Nr+1):
        for j in range(1,Nz+1):
            z_tau_a    [j-1,i-1] = tau_a    [i-1,j-1]
            z_tau_b1   [j-1,i-1] = tau_b1   [i-1,j-1]
            z_tau_b2   [j-1,i-1] = tau_b2   [i-1,j-1]
            z_u_z_a_c  [j-1,i-1] = u_z_a_c  [i-1,j-1]
            z_u_r_a_c  [j-1,i-1] = u_r_a_c  [i-1,j-1]
            z_u_z_a_d1 [j-1,i-1] = u_z_a_d1 [i-1,j-1]
            z_u_r_a_d1 [j-1,i-1] = u_r_a_d1 [i-1,j-1]
            z_u_z_a_d2 [j-1,i-1] = u_z_a_d2 [i-1,j-1]
            z_u_r_a_d2 [j-1,i-1] = u_r_a_d2 [i-1,j-1]
            z_u_z_b1_c [j-1,i-1] = u_z_b1_c [i-1,j-1]
            z_u_r_b1_c [j-1,i-1] = u_r_b1_c [i-1,j-1]
            z_u_z_b1_d1[j-1,i-1] = u_z_b1_d1[i-1,j-1]
            z_u_r_b1_d1[j-1,i-1] = u_r_b1_d1[i-1,j-1]
            z_u_z_b1_d2[j-1,i-1] = u_z_b1_d2[i-1,j-1]
            z_u_r_b1_d2[j-1,i-1] = u_r_b1_d2[i-1,j-1]
            z_u_z_b2_c [j-1,i-1] = u_z_b2_c [i-1,j-1]
            z_u_r_b2_c [j-1,i-1] = u_r_b2_c [i-1,j-1]
            z_u_z_b2_d1[j-1,i-1] = u_z_b2_d1[i-1,j-1]
            z_u_r_b2_d1[j-1,i-1] = u_r_b2_d1[i-1,j-1]
            z_u_z_b2_d2[j-1,i-1] = u_z_b2_d2[i-1,j-1]
            z_u_r_b2_d2[j-1,i-1] = u_r_b2_d2[i-1,j-1]
            z_u_u_a_c  [j-1,i-1] = mp.sqrt( u_r_a_c  [i-1,j-1] ** 2 + u_z_a_c  [i-1,j-1] ** 2 )
            z_u_u_a_d1 [j-1,i-1] = mp.sqrt( u_r_a_d1 [i-1,j-1] ** 2 + u_z_a_d1 [i-1,j-1] ** 2 )
            z_u_u_a_d2 [j-1,i-1] = mp.sqrt( u_r_a_d2 [i-1,j-1] ** 2 + u_z_a_d2 [i-1,j-1] ** 2 )
            z_u_u_b1_c [j-1,i-1] = mp.sqrt( u_r_b1_c [i-1,j-1] ** 2 + u_z_b1_c [i-1,j-1] ** 2 )
            z_u_u_b1_d1[j-1,i-1] = mp.sqrt( u_r_b1_d1[i-1,j-1] ** 2 + u_z_b1_d1[i-1,j-1] ** 2 )
            z_u_u_b1_d2[j-1,i-1] = mp.sqrt( u_r_b1_d2[i-1,j-1] ** 2 + u_z_b1_d2[i-1,j-1] ** 2 )
            z_u_u_b2_c [j-1,i-1] = mp.sqrt( u_r_b2_c [i-1,j-1] ** 2 + u_z_b2_c [i-1,j-1] ** 2 )
            z_u_u_b2_d1[j-1,i-1] = mp.sqrt( u_r_b2_d1[i-1,j-1] ** 2 + u_z_b2_d1[i-1,j-1] ** 2 ) 
            z_u_u_b2_d2[j-1,i-1] = mp.sqrt( u_r_b2_d2[i-1,j-1] ** 2 + u_z_b2_d2[i-1,j-1] ** 2 )



    fig, ax = plt.subplots(3, 4, layout=hdl_layout, figsize=(12, 6), dpi=300)            
    fig.suptitle(\
"non-dimensional flow field (w/o delta_Tx delta_uy) \
for fluid domain in (tilde{r},tilde{z})-space"\
)
    panel_tau_a   = ax[0,0].set_title('tau_a  ')
    panel_tau_b1  = ax[1,0].set_title('tau_b1 ')
    panel_tau_b2  = ax[2,0].set_title('tau_b2 ')
    panel_u_a_c   = ax[0,1].set_title('u_a_c  ')
    panel_u_b1_c  = ax[1,1].set_title('u_b1_c ')
    panel_u_b2_c  = ax[2,1].set_title('u_b2_c ')
    panel_u_a_d1  = ax[0,2].set_title('u_a_d1 ')
    panel_u_b1_d1 = ax[1,2].set_title('u_b1_d1')
    panel_u_b2_d1 = ax[2,2].set_title('u_b2_d1')
    panel_u_a_d2  = ax[0,3].set_title('u_a_d2 ')
    panel_u_b1_d2 = ax[1,3].set_title('u_b1_d2')
    panel_u_b2_d2 = ax[2,3].set_title('u_b2_d2')
    panel_tau_a   = ax[0,0].contourf(x,y,z_tau_a    ,cmap='RdBu_r',levels=ct_lv)
    panel_tau_b1  = ax[1,0].contourf(x,y,z_tau_b1   ,cmap='RdBu_r',levels=ct_lv)
    panel_tau_b2  = ax[2,0].contourf(x,y,z_tau_b2   ,cmap='RdBu_r',levels=ct_lv)
    panel_u_a_c   = ax[0,1].contourf(x,y,z_u_u_a_c  ,cmap='RdBu_r',levels=ct_lv)
    panel_u_b1_c  = ax[1,1].contourf(x,y,z_u_u_b1_c ,cmap='RdBu_r',levels=ct_lv)
    panel_u_b2_c  = ax[2,1].contourf(x,y,z_u_u_b2_c ,cmap='RdBu_r',levels=ct_lv)
    panel_u_a_d1  = ax[0,2].contourf(x,y,z_u_u_a_d1 ,cmap='RdBu_r',levels=ct_lv)
    panel_u_b1_d1 = ax[1,2].contourf(x,y,z_u_u_b1_d1,cmap='RdBu_r',levels=ct_lv)
    panel_u_b2_d1 = ax[2,2].contourf(x,y,z_u_u_b2_d1,cmap='RdBu_r',levels=ct_lv)
    panel_u_a_d2  = ax[0,3].contourf(x,y,z_u_u_a_d2 ,cmap='RdBu_r',levels=ct_lv)
    panel_u_b1_d2 = ax[1,3].contourf(x,y,z_u_u_b1_d2,cmap='RdBu_r',levels=ct_lv)
    panel_u_b2_d2 = ax[2,3].contourf(x,y,z_u_u_b2_d2,cmap='RdBu_r',levels=ct_lv) 

    fig.colorbar(panel_tau_a  , ax=ax[0,0], shrink = clb_shk)
    fig.colorbar(panel_tau_b1 , ax=ax[1,0], shrink = clb_shk)
    fig.colorbar(panel_tau_b2 , ax=ax[2,0], shrink = clb_shk)
    fig.colorbar(panel_u_a_c  , ax=ax[0,1], shrink = clb_shk)
    fig.colorbar(panel_u_b1_c , ax=ax[1,1], shrink = clb_shk)
    fig.colorbar(panel_u_b2_c , ax=ax[2,1], shrink = clb_shk)
    fig.colorbar(panel_u_a_d1 , ax=ax[0,2], shrink = clb_shk)
    fig.colorbar(panel_u_b1_d1, ax=ax[1,2], shrink = clb_shk)
    fig.colorbar(panel_u_b2_d1, ax=ax[2,2], shrink = clb_shk)
    fig.colorbar(panel_u_a_d2 , ax=ax[0,3], shrink = clb_shk)
    fig.colorbar(panel_u_b1_d2, ax=ax[1,3], shrink = clb_shk)
    fig.colorbar(panel_u_b2_d2, ax=ax[2,3], shrink = clb_shk)
    
    panel_tau_a   = ax[0,0].contour (x,y,z_tau_a ,linewidths=0.5,colors='k',levels=ct_lv)
    panel_tau_b1  = ax[1,0].contour (x,y,z_tau_b1,linewidths=0.5,colors='k',levels=ct_lv)
    panel_tau_b2  = ax[2,0].contour (x,y,z_tau_b2,linewidths=0.5,colors='k',levels=ct_lv)
    panel_u_a_c   = ax[0,1].quiver  (x,y,z_u_r_a_c  ,z_u_z_a_c  )
    panel_u_b1_c  = ax[1,1].quiver  (x,y,z_u_r_b1_c ,z_u_z_b1_c )
    panel_u_b2_c  = ax[2,1].quiver  (x,y,z_u_r_b2_c ,z_u_z_b2_c )
    panel_u_a_d1  = ax[0,2].quiver  (x,y,z_u_r_a_d1 ,z_u_z_a_d1 )
    panel_u_b1_d1 = ax[1,2].quiver  (x,y,z_u_r_b1_d1,z_u_z_b1_d1)
    panel_u_b2_d1 = ax[2,2].quiver  (x,y,z_u_r_b2_d1,z_u_z_b2_d1)
    panel_u_a_d2  = ax[0,3].quiver  (x,y,z_u_r_a_d2 ,z_u_z_a_d2 )
    panel_u_b1_d2 = ax[1,3].quiver  (x,y,z_u_r_b1_d2,z_u_z_b1_d2)
    panel_u_b2_d2 = ax[2,3].quiver  (x,y,z_u_r_b2_d2,z_u_z_b2_d2)

    for ab in range(3):
        ax[ab,0].set_aspect(hdl_aspect)
        for cd in range(1,4):
            ax[ab,cd].set_aspect(hdl_aspect)
    if display_macro_val_on_console == 'n':        
        plt.savefig(fname+'_nondimensional-fluid-domain.png')
    plt.show()

def plot_result() -> None:
    
    #whole temperature field
    #this is skipped in magnification mode
    if search_s_max == 'n' and search_Nm == 'n':
        print('>> [PLOT] dimensional quantities')  

    if magnify != 'y':
        for j in range(1,Nz_1+1): 
            tz_all[j-1] = tz_1[j-1]
        for j in range(1,Nz+1): 
            tz_all[j-1+Nz_1] = tz[j-1] 
        for j in range(1,Nz_2+1): 
            tz_all[j-1+Nz+Nz_1] = tz_2[j-1]
        x, y = np.meshgrid(tr * w_0, tz_all * w_0) 
        for i in range(1,Nr+1):
            for j in range(1,Nz_1+1):
                arg = j-1
                z_dT_all[arg,i-1] = dT_1[i-1,j-1] 
                z_Tr_all[arg,i-1] = mp.mpf('0')
                z_Tz_all[arg,i-1] = mp.mpf('0')
                z_TT_all[arg,i-1] = mp.mpf('0')
                z_vr_all[arg,i-1] = mp.mpf('0')
                z_vz_all[arg,i-1] = mp.mpf('0')
                z_vv_all[arg,i-1] = mp.mpf('0')
            for j in range(1,Nz+1):
                arg = j-1 + Nz_1
                z_dT_all[arg,i-1] = dT[i-1,j-1] 
                z_Tr_all[arg,i-1] = Tr[i-1,j-1]  
                z_Tz_all[arg,i-1] = Tz[i-1,j-1]  
                z_TT_all[arg,i-1] = mp.sqrt( Tr[i-1,j-1]*Tr[i-1,j-1] + Tz[i-1,j-1]*Tz[i-1,j-1] )
                z_vr_all[arg,i-1] = vr[i-1,j-1]  
                z_vz_all[arg,i-1] = vz[i-1,j-1]  
                z_vv_all[arg,i-1] = mp.sqrt( vr[i-1,j-1]*vr[i-1,j-1] + vz[i-1,j-1]*vz[i-1,j-1] )
            for j in range(1,Nz_2+1):
                arg = j-1 + Nz + Nz_1
                z_dT_all[arg,i-1] = dT_2[i-1,j-1] 
                z_Tr_all[arg,i-1] = mp.mpf('0')
                z_Tz_all[arg,i-1] = mp.mpf('0')
                z_TT_all[arg,i-1] = mp.mpf('0')
                z_vr_all[arg,i-1] = mp.mpf('0')
                z_vz_all[arg,i-1] = mp.mpf('0')
                z_vv_all[arg,i-1] = mp.mpf('0')
    else:
        x, y = np.meshgrid(tr * w_0, tz * w_0) 
        for i in range(1,Nr+1):
            for j in range(1,Nz+1):
                z_dT[j-1,i-1] = dT[i-1,j-1] 
                z_Tr[j-1,i-1] = Tr[i-1,j-1]  
                z_Tz[j-1,i-1] = Tz[i-1,j-1]  
                z_TT[j-1,i-1] = mp.sqrt( Tr[i-1,j-1]*Tr[i-1,j-1] + Tz[i-1,j-1]*Tz[i-1,j-1] )
                z_vr[j-1,i-1] = vr[i-1,j-1]  
                z_vz[j-1,i-1] = vz[i-1,j-1]  
                z_vv[j-1,i-1] = mp.sqrt( vr[i-1,j-1]*vr[i-1,j-1] + vz[i-1,j-1]*vz[i-1,j-1] )
                
    fig, axsd = plt.subplots(1, 3, layout=hdl_layout, figsize=(12,8), dpi=300)            
    fig.suptitle(\
"temperature increase, temperature gradient, flow field \
in (r,z)-space \n case="+\
mp.nstr(case)+", Nm="+str(Nm)+", s_max0="+str(s_max0)+" (s_max="+mp.nstr(s_max,3)+")", 
)
    if magnify != 'y':
        panel_dT  = axsd[0].contour (x,y,z_dT_all,linewidths=0.5,colors='k',levels=20)
        panel_dT  = axsd[0].contourf(x,y,z_dT_all,cmap='RdBu_r',levels=20)
        panel_TT  = axsd[1].contour (x,y,z_TT_all,linewidths=0.5,colors='k',levels=20)
        panel_TT  = axsd[1].contourf(x,y,z_TT_all,cmap='RdBu_r',levels=20)
        panel_vv  = axsd[2].contour (x,y,z_vv_all,linewidths=0.5,colors='k',levels=20)
        panel_vv  = axsd[2].contourf(x,y,z_vv_all,cmap='RdBu_r',levels=20)
    else:
        panel_dT  = axsd[0].contour (x,y,z_dT,linewidths=0.5,colors='k',levels=20)
        panel_dT  = axsd[0].contourf(x,y,z_dT,cmap='RdBu_r',levels=20)
        panel_TT  = axsd[1].contour (x,y,z_TT,linewidths=0.5,colors='k',levels=20)
        panel_TT  = axsd[1].contourf(x,y,z_TT,cmap='RdBu_r',levels=20)
        panel_vv  = axsd[2].contour (x,y,z_vv,linewidths=0.5,colors='k',levels=20)
        panel_vv  = axsd[2].contourf(x,y,z_vv,cmap='RdBu_r',levels=20)
    for ab in range(3):
        if magnify != 'y':
            axsd[ab].axhline(0,color='black',ls='--', linewidth=1.0)
            axsd[ab].axhline(H,color='black',ls='--', linewidth=1.0)
        axsd[ab].set_aspect(hdl_aspect)
    fig.colorbar(panel_dT, ax=axsd[0], shrink = clb_shk)
    fig.colorbar(panel_TT, ax=axsd[1], shrink = clb_shk)
    fig.colorbar(panel_vv, ax=axsd[2], shrink = clb_shk)
    axsd[0].set_title('T-T0 [K]')
    axsd[1].set_title('grad T [K/m]')
    axsd[2].set_title('v [m/s]')
    if magnify != 'y':
        panel_TT = axsd[1].quiver(x,y,z_Tr_all,z_Tz_all)
        panel_vv = axsd[2].quiver(x,y,z_vr_all,z_vz_all)
    else:
        panel_TT = axsd[1].quiver(x,y,z_Tr,z_Tz)
        panel_vv = axsd[2].quiver(x,y,z_vr,z_vz)
    if display_macro_val_on_console == 'n':    
        if magnify == 'y':
            plt.savefig(fname+'_dimensional_quantities_magnify.png')
        else:
            plt.savefig(fname+'_dimensional_quantities.png')
    plt.show()


def define_arrays() -> None: 
    global sJN
    global tau_a, tau_b1, tau_b2
    global tau_a_r, tau_b1_r, tau_b2_r
    global tau_a_z, tau_b1_z, tau_b2_z
    global tau_a_1, tau_b1_1, tau_b2_1
    global tau_a_2, tau_b1_2, tau_b2_2
    global bta, bta_r, bta_z
    global btb1, btb1_r, btb1_z
    global btb2, btb2_r, btb2_z
    global bta_1, btb1_1, btb2_1
    global bta_2, btb1_2, btb2_2
    global coef_Bm, coef_Bp #these will be used in "get_C"
    global coef_B1, coef_B2 
    global coef_C1p_c,  coef_C2p_c,  coef_C3p_c,  coef_C1m_c,  coef_C2m_c,  coef_C3m_c 
    global coef_C1p_d1, coef_C2p_d1, coef_C3p_d1, coef_C1m_d1, coef_C2m_d1, coef_C3m_d1
    global coef_C1p_d2, coef_C2p_d2, coef_C3p_d2, coef_C1m_d2, coef_C2m_d2, coef_C3m_d2
    global bu_z_a_c,   bu_r_a_c   
    global bu_z_a_d1,  bu_r_a_d1  
    global bu_z_a_d2,  bu_r_a_d2  
    global bu_z_b1_c,  bu_r_b1_c  
    global bu_z_b1_d1, bu_r_b1_d1 
    global bu_z_b1_d2, bu_r_b1_d2 
    global bu_z_b2_c,  bu_r_b2_c  
    global bu_z_b2_d1, bu_r_b2_d1 
    global bu_z_b2_d2, bu_r_b2_d2 
    global u_z_a_c,  u_z_b1_c,  u_z_b2_c
    global u_z_a_d1, u_z_b1_d1, u_z_b2_d1
    global u_z_a_d2, u_z_b1_d2, u_z_b2_d2
    global u_r_a_c,  u_r_b1_c,  u_r_b2_c
    global u_r_a_d1, u_r_b1_d1, u_r_b2_d1
    global u_r_a_d2, u_r_b1_d2, u_r_b2_d2
    global dT   
    global dT_1 
    global dT_2 
    global Tr   
    global Tz   
    global vr  
    global vz  
    global tz_all       
    global z_tau_a_all  
    global z_tau_b1_all 
    global z_tau_b2_all 
    global z_tau_a     
    global z_tau_b1    
    global z_tau_b2    
    global z_u_z_a_c   
    global z_u_r_a_c   
    global z_u_z_a_d1  
    global z_u_r_a_d1  
    global z_u_z_a_d2  
    global z_u_r_a_d2  
    global z_u_z_b1_c  
    global z_u_r_b1_c  
    global z_u_z_b1_d1 
    global z_u_r_b1_d1 
    global z_u_z_b1_d2 
    global z_u_r_b1_d2 
    global z_u_z_b2_c  
    global z_u_r_b2_c  
    global z_u_z_b2_d1 
    global z_u_r_b2_d1 
    global z_u_z_b2_d2 
    global z_u_r_b2_d2 
    global z_u_u_a_c   
    global z_u_u_a_d1  
    global z_u_u_a_d2  
    global z_u_u_b1_c  
    global z_u_u_b1_d1 
    global z_u_u_b1_d2 
    global z_u_u_b2_c  
    global z_u_u_b2_d1 
    global z_u_u_b2_d2 
    global z_dT_all, z_dT
    global z_Tr, z_Tz, z_TT
    global z_Tr_all, z_Tz_all, z_TT_all
    global z_vr_all, z_vz_all, z_vv_all
    global z_vr, z_vz, z_vv
    global chk_tau_a    
    global chk_tau_a_r  
    global chk_tau_a_z  
    global chk_tau_b1   
    global chk_tau_b1_r 
    global chk_tau_b1_z 
    global chk_tau_b2   
    global chk_tau_b2_r 
    global chk_tau_b2_z 
    global chk_u_z_a_c  
    global chk_u_r_a_c  
    global chk_u_z_a_d1 
    global chk_u_r_a_d1 
    global chk_u_z_a_d2 
    global chk_u_r_a_d2 
    global chk_u_z_b1_c 
    global chk_u_r_b1_c 
    global chk_u_z_b1_d1
    global chk_u_r_b1_d1
    global chk_u_z_b1_d2
    global chk_u_r_b1_d2
    global chk_u_z_b2_c 
    global chk_u_r_b2_c 
    global chk_u_z_b2_d1
    global chk_u_r_b2_d1
    global chk_u_z_b2_d2
    global chk_u_r_b2_d2
    global dT_old
    global Tr_old
    global Tz_old
    global vr_old
    global vz_old
    

        
    sJN      = mp.matrix(Nm,2)
    tau_a    = mp.matrix(Nr,Nz);   tau_a_r  = mp.matrix(Nr,Nz);   tau_a_z  = mp.matrix(Nr,Nz)
    tau_b1   = mp.matrix(Nr,Nz);   tau_b1_r = mp.matrix(Nr,Nz);   tau_b1_z = mp.matrix(Nr,Nz)
    tau_b2   = mp.matrix(Nr,Nz);   tau_b2_r = mp.matrix(Nr,Nz);   tau_b2_z = mp.matrix(Nr,Nz)
    tau_a_1  = mp.matrix(Nr,Nz_1); tau_b1_1 = mp.matrix(Nr,Nz_1); tau_b2_1 = mp.matrix(Nr,Nz_1)
    tau_a_2  = mp.matrix(Nr,Nz_2); tau_b1_2 = mp.matrix(Nr,Nz_2); tau_b2_2 = mp.matrix(Nr,Nz_2)
    bta      = mp.matrix(Nm,Nz);   bta_r    = mp.matrix(Nm,Nz);   bta_z    = mp.matrix(Nm,Nz)
    btb1     = mp.matrix(Nm,Nz);   btb1_r   = mp.matrix(Nm,Nz);   btb1_z   = mp.matrix(Nm,Nz)
    btb2     = mp.matrix(Nm,Nz);   btb2_r   = mp.matrix(Nm,Nz);   btb2_z   = mp.matrix(Nm,Nz)
    bta_1    = mp.matrix(Nm,Nz_1); btb1_1   = mp.matrix(Nm,Nz_1); btb2_1   = mp.matrix(Nm,Nz_1)
    bta_2    = mp.matrix(Nm,Nz_2); btb1_2   = mp.matrix(Nm,Nz_2); btb2_2   = mp.matrix(Nm,Nz_2)
    coef_Bm  = mp.matrix(Nm,3);    coef_Bp  = mp.matrix(Nm,3)
    coef_B1  = mp.matrix(Nm,3);    coef_B2  = mp.matrix(Nm,3)
    coef_C1p_c  = mp.matrix(Nm,3); coef_C2p_c  = mp.matrix(Nm,3); coef_C3p_c  = mp.matrix(Nm,3)
    coef_C1m_c  = mp.matrix(Nm,3); coef_C2m_c  = mp.matrix(Nm,3); coef_C3m_c  = mp.matrix(Nm,3)
    coef_C1p_d1 = mp.matrix(Nm,3); coef_C2p_d1 = mp.matrix(Nm,3); coef_C3p_d1 = mp.matrix(Nm,3)
    coef_C1m_d1 = mp.matrix(Nm,3); coef_C2m_d1 = mp.matrix(Nm,3); coef_C3m_d1 = mp.matrix(Nm,3)
    coef_C1p_d2 = mp.matrix(Nm,3); coef_C2p_d2 = mp.matrix(Nm,3); coef_C3p_d2 = mp.matrix(Nm,3)
    coef_C1m_d2 = mp.matrix(Nm,3); coef_C2m_d2 = mp.matrix(Nm,3); coef_C3m_d2 = mp.matrix(Nm,3)
    bu_z_a_c   = mp.matrix(Nm,Nz); bu_r_a_c   = mp.matrix(Nm,Nz)
    bu_z_a_d1  = mp.matrix(Nm,Nz); bu_r_a_d1  = mp.matrix(Nm,Nz)
    bu_z_a_d2  = mp.matrix(Nm,Nz); bu_r_a_d2  = mp.matrix(Nm,Nz)
    bu_z_b1_c  = mp.matrix(Nm,Nz); bu_r_b1_c  = mp.matrix(Nm,Nz)
    bu_z_b1_d1 = mp.matrix(Nm,Nz); bu_r_b1_d1 = mp.matrix(Nm,Nz)
    bu_z_b1_d2 = mp.matrix(Nm,Nz); bu_r_b1_d2 = mp.matrix(Nm,Nz)
    bu_z_b2_c  = mp.matrix(Nm,Nz); bu_r_b2_c  = mp.matrix(Nm,Nz)
    bu_z_b2_d1 = mp.matrix(Nm,Nz); bu_r_b2_d1 = mp.matrix(Nm,Nz)
    bu_z_b2_d2 = mp.matrix(Nm,Nz); bu_r_b2_d2 = mp.matrix(Nm,Nz)
    u_z_a_c  = mp.matrix(Nr,Nz); u_z_b1_c  = mp.matrix(Nr,Nz); u_z_b2_c  = mp.matrix(Nr,Nz)
    u_z_a_d1 = mp.matrix(Nr,Nz); u_z_b1_d1 = mp.matrix(Nr,Nz); u_z_b2_d1 = mp.matrix(Nr,Nz)
    u_z_a_d2 = mp.matrix(Nr,Nz); u_z_b1_d2 = mp.matrix(Nr,Nz); u_z_b2_d2 = mp.matrix(Nr,Nz)
    u_r_a_c  = mp.matrix(Nr,Nz); u_r_b1_c  = mp.matrix(Nr,Nz); u_r_b2_c  = mp.matrix(Nr,Nz)
    u_r_a_d1 = mp.matrix(Nr,Nz); u_r_b1_d1 = mp.matrix(Nr,Nz); u_r_b2_d1 = mp.matrix(Nr,Nz)
    u_r_a_d2 = mp.matrix(Nr,Nz); u_r_b1_d2 = mp.matrix(Nr,Nz); u_r_b2_d2 = mp.matrix(Nr,Nz)
    
    #[dimensional]
    dT   = np.zeros((Nr,Nz))
    dT_1 = np.zeros((Nr,Nz_1))
    dT_2 = np.zeros((Nr,Nz_2))
    Tr   = np.zeros((Nr,Nz))
    Tz   = np.zeros((Nr,Nz))
    vr   = np.zeros((Nr,Nz))
    vz   = np.zeros((Nr,Nz))
    #[dimensional]
    if pp == 0:
        dT_old  = np.zeros((Nr,Nz,pmax))
        Tr_old  = np.zeros((Nr,Nz,pmax))
        Tz_old  = np.zeros((Nr,Nz,pmax))
        vr_old  = np.zeros((Nr,Nz,pmax))
        vz_old  = np.zeros((Nr,Nz,pmax))
    
    #[PLOT]
    #nd temperature
    tz_all       = np.zeros(Nz+Nz_1+Nz_2)
    z_tau_a_all  = np.zeros((Nz+Nz_1+Nz_2,Nr)) 
    z_tau_b1_all = np.zeros((Nz+Nz_1+Nz_2,Nr)) 
    z_tau_b2_all = np.zeros((Nz+Nz_1+Nz_2,Nr)) 
    #nd fluid domain
    z_tau_a     = np.zeros((Nz,Nr))
    z_tau_b1    = np.zeros((Nz,Nr)) 
    z_tau_b2    = np.zeros((Nz,Nr)) 
    z_u_z_a_c   = np.zeros((Nz,Nr))
    z_u_r_a_c   = np.zeros((Nz,Nr))
    z_u_z_a_d1  = np.zeros((Nz,Nr))
    z_u_r_a_d1  = np.zeros((Nz,Nr))
    z_u_z_a_d2  = np.zeros((Nz,Nr))
    z_u_r_a_d2  = np.zeros((Nz,Nr))
    z_u_z_b1_c  = np.zeros((Nz,Nr))
    z_u_r_b1_c  = np.zeros((Nz,Nr))
    z_u_z_b1_d1 = np.zeros((Nz,Nr))
    z_u_r_b1_d1 = np.zeros((Nz,Nr))
    z_u_z_b1_d2 = np.zeros((Nz,Nr))
    z_u_r_b1_d2 = np.zeros((Nz,Nr))
    z_u_z_b2_c  = np.zeros((Nz,Nr))
    z_u_r_b2_c  = np.zeros((Nz,Nr))
    z_u_z_b2_d1 = np.zeros((Nz,Nr))
    z_u_r_b2_d1 = np.zeros((Nz,Nr))
    z_u_z_b2_d2 = np.zeros((Nz,Nr))
    z_u_r_b2_d2 = np.zeros((Nz,Nr))
    z_u_u_a_c   = np.zeros((Nz,Nr))
    z_u_u_a_d1  = np.zeros((Nz,Nr))
    z_u_u_a_d2  = np.zeros((Nz,Nr))
    z_u_u_b1_c  = np.zeros((Nz,Nr))
    z_u_u_b1_d1 = np.zeros((Nz,Nr))
    z_u_u_b1_d2 = np.zeros((Nz,Nr))
    z_u_u_b2_c  = np.zeros((Nz,Nr))
    z_u_u_b2_d1 = np.zeros((Nz,Nr))
    z_u_u_b2_d2 = np.zeros((Nz,Nr))
    
    #dimensional temperature
    z_dT_all = np.zeros((Nz+Nz_1+Nz_2,Nr))
    z_Tr_all = np.zeros((Nz+Nz_1+Nz_2,Nr))
    z_Tz_all = np.zeros((Nz+Nz_1+Nz_2,Nr))
    z_TT_all = np.zeros((Nz+Nz_1+Nz_2,Nr))
    z_dT     = np.zeros((Nz,Nr))
    z_Tr     = np.zeros((Nz,Nr))
    z_Tz     = np.zeros((Nz,Nr))
    z_TT     = np.zeros((Nz,Nr))
    z_vr_all = np.zeros((Nz+Nz_1+Nz_2,Nr))
    z_vz_all = np.zeros((Nz+Nz_1+Nz_2,Nr))
    z_vv_all = np.zeros((Nz+Nz_1+Nz_2,Nr))
    z_vr     = np.zeros((Nz,Nr))
    z_vz     = np.zeros((Nz,Nr))
    z_vv     = np.zeros((Nz,Nr))
    
    chk_tau_a     = mp.matrix(Nr,Nz)
    chk_tau_a_r   = mp.matrix(Nr,Nz)
    chk_tau_a_z   = mp.matrix(Nr,Nz)
    chk_tau_b1    = mp.matrix(Nr,Nz)
    chk_tau_b1_r  = mp.matrix(Nr,Nz)
    chk_tau_b1_z  = mp.matrix(Nr,Nz)
    chk_tau_b2    = mp.matrix(Nr,Nz)
    chk_tau_b2_r  = mp.matrix(Nr,Nz)
    chk_tau_b2_z  = mp.matrix(Nr,Nz)
    chk_u_z_a_c   = mp.matrix(Nr,Nz)
    chk_u_r_a_c   = mp.matrix(Nr,Nz)
    chk_u_z_a_d1  = mp.matrix(Nr,Nz)
    chk_u_r_a_d1  = mp.matrix(Nr,Nz)
    chk_u_z_a_d2  = mp.matrix(Nr,Nz)
    chk_u_r_a_d2  = mp.matrix(Nr,Nz)
    chk_u_z_b1_c  = mp.matrix(Nr,Nz)
    chk_u_r_b1_c  = mp.matrix(Nr,Nz)
    chk_u_z_b1_d1 = mp.matrix(Nr,Nz)
    chk_u_r_b1_d1 = mp.matrix(Nr,Nz)
    chk_u_z_b1_d2 = mp.matrix(Nr,Nz)
    chk_u_r_b1_d2 = mp.matrix(Nr,Nz)
    chk_u_z_b2_c  = mp.matrix(Nr,Nz)
    chk_u_r_b2_c  = mp.matrix(Nr,Nz)
    chk_u_z_b2_d1 = mp.matrix(Nr,Nz)
    chk_u_r_b2_d1 = mp.matrix(Nr,Nz)
    chk_u_z_b2_d2 = mp.matrix(Nr,Nz)
    chk_u_r_b2_d2 = mp.matrix(Nr,Nz)

def compute_mac():
    
    define_arrays()
    
    if search_s_max == 'n' and search_Nm == 'n':
        print('>> [COMPUTING] coefficient B')
    for m in range(1,Nm+1): #1, ..., Nm
        X = mp.matrix(4,3)
        X = get_B( s[m-1] )
        for ab in range(3):
            coef_Bm[m-1,ab] = X[0,ab]
            coef_Bp[m-1,ab] = X[1,ab]
            coef_B1[m-1,ab] = X[2,ab]
            coef_B2[m-1,ab] = X[3,ab]
            
    if search_s_max == 'n' and search_Nm == 'n':            
        print('>> [COMPUTING] coefficient C')
    for m in range(1,Nm+1): #1, ..., Nm
        Yc  = mp.matrix(6,3)
        Yd1 = mp.matrix(6,3)
        Yd2 = mp.matrix(6,3)
        Yc, Yd1, Yd2 = get_C( m-1, s[m-1] )
        for ab in range(3):
            coef_C1p_c [m-1,ab] = Yc [0,ab]
            coef_C2p_c [m-1,ab] = Yc [1,ab]
            coef_C3p_c [m-1,ab] = Yc [2,ab]
            coef_C1m_c [m-1,ab] = Yc [3,ab]
            coef_C2m_c [m-1,ab] = Yc [4,ab]
            coef_C3m_c [m-1,ab] = Yc [5,ab]
            coef_C1p_d1[m-1,ab] = Yd1[0,ab]
            coef_C2p_d1[m-1,ab] = Yd1[1,ab]
            coef_C3p_d1[m-1,ab] = Yd1[2,ab]
            coef_C1m_d1[m-1,ab] = Yd1[3,ab]
            coef_C2m_d1[m-1,ab] = Yd1[4,ab]
            coef_C3m_d1[m-1,ab] = Yd1[5,ab]
            coef_C1p_d2[m-1,ab] = Yd2[0,ab]
            coef_C2p_d2[m-1,ab] = Yd2[1,ab]
            coef_C3p_d2[m-1,ab] = Yd2[2,ab]
            coef_C1m_d2[m-1,ab] = Yd2[3,ab]
            coef_C2m_d2[m-1,ab] = Yd2[4,ab]
            coef_C3m_d2[m-1,ab] = Yd2[5,ab]
        #*** for debug use    
        # ab = 2
        # if m == 3:
        #     print(mp.nstr(coef_C1p_d2[m-1,ab]),\
        #           mp.nstr(coef_C2p_d2[m-1,ab]),\
        #           mp.nstr(coef_C3p_d2[m-1,ab]),\
        #           mp.nstr(coef_C1m_d2[m-1,ab]),\
        #           mp.nstr(coef_C2m_d2[m-1,ab]),\
        #           mp.nstr(coef_C3m_d2[m-1,ab]) \
        #           )

    if search_s_max == 'n' and search_Nm == 'n':                                              
        print('>> [COMPUTING] transformed functions')        
        print('   --- bt (bar tau) ---')
    start_bt = time.time()
    for j in range(1,Nz+1):
        
        if Nz > 10: 
            div_j = 5
        else:
            div_j = 2
        if j%div_j == 0:
            end_bt = time.time()
            time_diff = end_bt - start_bt
            if search_s_max == 'n' and search_Nm == 'n':
                print('   ',"{:3}".format(int(j/Nz*100)),'% done (',"{:.2f}".format(time_diff),'sec)')
            
        for m in range(1,Nm+1):
            tausa  = compute_tausa ( s[m-1] )  
            tausb1 = compute_tausb1( s[m-1] )  
            tausb2 = compute_tausb2( s[m-1] )  
            Ezm = func_E( -s[m-1], tz[j-1] - tz_0 )
            Ezp = func_E(  s[m-1], tz[j-1] - tz_0 )
            Fzm = func_F( -s[m-1], tz[j-1] - tz_0 )
            Fzp = func_F(  s[m-1], tz[j-1] - tz_0 )
            Gzm = func_G( -s[m-1], tz[j-1] - tz_0 )
            Gzp = func_G(  s[m-1], tz[j-1] - tz_0 )
            auxa   =   ( coef_Bm[m-1,0] + Ezm       )*Fzm + ( coef_Bp[m-1,0] + Ezp       )*Fzp
            auxpa  = - ( coef_Bm[m-1,0] + Ezm - Gzm )*Fzm + ( coef_Bp[m-1,0] + Ezp - Gzp )*Fzp
            auxb1  =   ( coef_Bm[m-1,1]             )*Fzm + ( coef_Bp[m-1,1]             )*Fzp
            auxpb1 = - ( coef_Bm[m-1,1]             )*Fzm + ( coef_Bp[m-1,1]             )*Fzp
            auxb2  =   ( coef_Bm[m-1,2]             )*Fzm + ( coef_Bp[m-1,2]             )*Fzp
            auxpb2 = - ( coef_Bm[m-1,2]             )*Fzm + ( coef_Bp[m-1,2]             )*Fzp
            bta   [m-1,j-1] =            ( tausa  * auxa   )
            bta_r [m-1,j-1] = - s[m-1] * ( tausa  * auxa   )
            bta_z [m-1,j-1] =   s[m-1] * ( tausa  * auxpa  )
            btb1  [m-1,j-1] =            ( tausb1 * auxb1  )
            btb1_r[m-1,j-1] = - s[m-1] * ( tausb1 * auxb1  )
            btb1_z[m-1,j-1] =   s[m-1] * ( tausb1 * auxpb1 )
            btb2  [m-1,j-1] =            ( tausb2 * auxb2  )
            btb2_r[m-1,j-1] = - s[m-1] * ( tausb2 * auxb2  )
            btb2_z[m-1,j-1] =   s[m-1] * ( tausb2 * auxpb2 )
            if m == 1:
                bta [m-1,j-1] = mp.mpf('0')
                btb1[m-1,j-1] = mp.mpf('0')
                btb2[m-1,j-1] = mp.mpf('0')
    if search_s_max == 'n' and search_Nm == 'n':            
        print('   --- bu (bar ur and uz) ---')
    start_bu = time.time()
    for j in range(1,Nz+1):
        
        if Nz > 10: 
            div_j = 5
        else:
            div_j = 2
        if j%div_j == 0:
            end_bu = time.time()
            time_diff = end_bu - start_bu
            if search_s_max == 'n' and search_Nm == 'n':
                print('   ',"{:3}".format(int(j/Nz*100)),'% done (',"{:.2f}".format(time_diff),'sec)')
        
        #*** !!! caution !!! this computation needs btx(m,0) and btx(m,Nz) ***
        zz = tz[j-1] / tH
        for m in range(1,Nm+1):
            tausa  = compute_tausa ( s[m-1] )  
            tausb1 = compute_tausb1( s[m-1] )  
            tausb2 = compute_tausb2( s[m-1] )  
            Fzm = func_F( -s[m-1], tz[j-1] - tz_0 )
            Fzp = func_F(  s[m-1], tz[j-1] - tz_0 )
            Szp = func_S(  s[m-1], tz[j-1] - tz_0 )
            Szm = func_S( -s[m-1], tz[j-1] - tz_0 )
            Vzp = func_V(  s[m-1], tz[j-1] - tz_0 )
            Vzm = func_V( -s[m-1], tz[j-1] - tz_0 )
            phi1  = zz * ( mp.mpf('1') - zz ) 
            phi2  = zz
            phi3  = mp.mpf('1') - zz  
            psi1p =   s[m-1] * phi1 + ( mp.mpf('1') - mp.mpf('2') * zz )/tH
            psi2p =   s[m-1] * phi2 + mp.mpf('1')/tH  
            psi3p =   s[m-1] * phi3 - mp.mpf('1')/tH  
            psi1m = - s[m-1] * phi1 + ( mp.mpf('1') - mp.mpf('2') * zz )/tH
            psi2m = - s[m-1] * phi2 + mp.mpf('1')/tH  
            psi3m = - s[m-1] * phi3 - mp.mpf('1')/tH  
            for ab in range(3):
                for cd in range(3):
                    if cd == 0:
                        tC1p_z = coef_C1p_c[m-1,ab] * phi1
                        tC2p_z = coef_C2p_c[m-1,ab] * phi2
                        tC3p_z = coef_C3p_c[m-1,ab] * phi3
                        tC1m_z = coef_C1m_c[m-1,ab] * phi1
                        tC2m_z = coef_C2m_c[m-1,ab] * phi2
                        tC3m_z = coef_C3m_c[m-1,ab] * phi3
                        tC1p_r = coef_C1p_c[m-1,ab] * psi1p
                        tC2p_r = coef_C2p_c[m-1,ab] * psi2p
                        tC3p_r = coef_C3p_c[m-1,ab] * psi3p
                        tC1m_r = coef_C1m_c[m-1,ab] * psi1m
                        tC2m_r = coef_C2m_c[m-1,ab] * psi2m
                        tC3m_r = coef_C3m_c[m-1,ab] * psi3m
                    elif cd == 1:
                        tC1p_z = coef_C1p_d1[m-1,ab] * phi1
                        tC2p_z = coef_C2p_d1[m-1,ab] * phi2
                        tC3p_z = coef_C3p_d1[m-1,ab] * phi3
                        tC1m_z = coef_C1m_d1[m-1,ab] * phi1
                        tC2m_z = coef_C2m_d1[m-1,ab] * phi2
                        tC3m_z = coef_C3m_d1[m-1,ab] * phi3
                        tC1p_r = coef_C1p_d1[m-1,ab] * psi1p
                        tC2p_r = coef_C2p_d1[m-1,ab] * psi2p
                        tC3p_r = coef_C3p_d1[m-1,ab] * psi3p
                        tC1m_r = coef_C1m_d1[m-1,ab] * psi1m
                        tC2m_r = coef_C2m_d1[m-1,ab] * psi2m
                        tC3m_r = coef_C3m_d1[m-1,ab] * psi3m
                    elif cd == 2:
                        tC1p_z = coef_C1p_d2[m-1,ab] * phi1
                        tC2p_z = coef_C2p_d2[m-1,ab] * phi2
                        tC3p_z = coef_C3p_d2[m-1,ab] * phi3
                        tC1m_z = coef_C1m_d2[m-1,ab] * phi1
                        tC2m_z = coef_C2m_d2[m-1,ab] * phi2
                        tC3m_z = coef_C3m_d2[m-1,ab] * phi3
                        tC1p_r = coef_C1p_d2[m-1,ab] * psi1p
                        tC2p_r = coef_C2p_d2[m-1,ab] * psi2p
                        tC3p_r = coef_C3p_d2[m-1,ab] * psi3p
                        tC1m_r = coef_C1m_d2[m-1,ab] * psi1m
                        tC2m_r = coef_C2m_d2[m-1,ab] * psi2m
                        tC3m_r = coef_C3m_d2[m-1,ab] * psi3m
                    if ab == 0 and cd == 0: 
                    #case of (a) fluid heating and (c) thermal convection
                        zeroone = mp.mpf('1')
                    else:
                        zeroone = mp.mpf('0')
                    tC_z = tC1p_z + tC2p_z + tC3p_z + Szp * zeroone
                    tD_z = tC1m_z + tC2m_z + tC3m_z + Szm * zeroone
                    tC_r = tC1p_r + tC2p_r + tC3p_r + Vzp * zeroone
                    tD_r = tC1m_r + tC2m_r + tC3m_r + Vzm * zeroone
                    bu_z = tC_z * Fzp + tD_z * Fzm
                    bu_r = tC_r * Fzp + tD_r * Fzm
                    if m == 1:
                        bu_z = mp.mpf('0')
                        bu_r = mp.mpf('0')
                    if cd == 0:
                        if ab == 0:
                            us = tausa / ( mp.mpf('2') * s[m-1] * s[m-1] )  
                            bu_z_a_c[m-1,j-1] = us * bu_z
                            bu_r_a_c[m-1,j-1] = us * bu_r * ( - mp.mpf('1')/s[m-1] )
                        elif ab == 1:
                            us = tausb1/ ( mp.mpf('2') * s[m-1] * s[m-1] )  
                            bu_z_b1_c[m-1,j-1] = us * bu_z
                            bu_r_b1_c[m-1,j-1] = us * bu_r * ( - mp.mpf('1')/s[m-1] )
                        elif ab == 2:
                            us = tausb2/ ( mp.mpf('2') * s[m-1] * s[m-1] )  
                            bu_z_b2_c[m-1,j-1] = us * bu_z
                            bu_r_b2_c[m-1,j-1] = us * bu_r * ( - mp.mpf('1')/s[m-1] )
                    elif cd == 1: 
                        if ab == 0:
                            us = s[m-1] * bta[m-1,0]
                            bu_z_a_d1[m-1,j-1] = us * bu_z
                            bu_r_a_d1[m-1,j-1] = us * bu_r * ( - mp.mpf('1')/s[m-1] )
                        elif ab == 1:
                            us = s[m-1] * btb1[m-1,0]
                            bu_z_b1_d1[m-1,j-1] = us * bu_z
                            bu_r_b1_d1[m-1,j-1] = us * bu_r * ( - mp.mpf('1')/s[m-1] )
                        elif ab == 2:
                            us = s[m-1] * btb2[m-1,0]
                            bu_z_b2_d1[m-1,j-1] = us * bu_z
                            bu_r_b2_d1[m-1,j-1] = us * bu_r * ( - mp.mpf('1')/s[m-1] )
                    elif cd == 2: 
                        if ab == 0:
                            us = s[m-1] * bta[m-1,Nz-1]
                            bu_z_a_d2[m-1,j-1] = us * bu_z
                            bu_r_a_d2[m-1,j-1] = us * bu_r * ( - mp.mpf('1')/s[m-1] )
                        elif ab == 1:
                            us = s[m-1] * btb1[m-1,Nz-1]
                            bu_z_b1_d2[m-1,j-1] = us * bu_z
                            bu_r_b1_d2[m-1,j-1] = us * bu_r * ( - mp.mpf('1')/s[m-1] )
                        elif ab == 2:
                            us = s[m-1] * btb2[m-1,Nz-1]
                            bu_z_b2_d2[m-1,j-1] = us * bu_z
                            bu_r_b2_d2[m-1,j-1] = us * bu_r * ( - mp.mpf('1')/s[m-1] )
    
    if search_s_max == 'n' and search_Nm == 'n':
        print('   --- bt_1 (bar tau_1) ---')
    for j in range(1,Nz_1+1):
        for m in range(1,Nm+1):
            tausa  = compute_tausa ( s[m-1] )  
            tausb1 = compute_tausb1( s[m-1] )  
            tausb2 = compute_tausb2( s[m-1] )  
            Fzp = func_F(  s[m-1], tz_1[j-1] + tH_1 ) 
            Fzm = func_F( -s[m-1], tz_1[j-1] + tH_1 ) 
            auxa  = coef_B1[m-1,0] * ( Fzp - Fzm )
            auxb1 = coef_B1[m-1,1] * ( Fzp - Fzm )
            auxb2 = coef_B1[m-1,2] * ( Fzp - Fzm )
            bta_1 [m-1,j-1] = tausa  * auxa
            btb1_1[m-1,j-1] = tausb1 * auxb1
            btb2_1[m-1,j-1] = tausb2 * auxb2
    if search_s_max == 'n' and search_Nm == 'n':  
        print('   --- bt_2 (bar tau_2) ---')              
    for j in range(1,Nz_2+1):
        for m in range(1,Nm+1):
            tausa  = compute_tausa ( s[m-1] )  
            tausb1 = compute_tausb1( s[m-1] )  
            tausb2 = compute_tausb2( s[m-1] )  
            Fzp = func_F(  s[m-1], tz_2[j-1] - tH - tH_2 ) 
            Fzm = func_F( -s[m-1], tz_2[j-1] - tH - tH_2 ) 
            auxa  = coef_B2[m-1,0] * ( Fzp - Fzm )
            auxb1 = coef_B2[m-1,1] * ( Fzp - Fzm )
            auxb2 = coef_B2[m-1,2] * ( Fzp - Fzm )
            bta_2 [m-1,j-1] = tausa  * auxa
            btb1_2[m-1,j-1] = tausb1 * auxb1
            btb2_2[m-1,j-1] = tausb2 * auxb2
    
    #--- memo ---
    #BesselJ(0,x) = 1-(1/4)*x^2+(1/64)*x^4+O(x^6) (x<<1)
    #BesselJ(1,x) = (1/2)*x-(1/16)*x^3+(1/384)*x^5+O(x^7) (x<<1)     
    start_inverse_Hankel = time.time()
    if search_s_max == 'n' and search_Nm == 'n':
        print('>> [COMPUTING] inverse Hankel transform')      
    for i in range(1,Nr+1):
        
        if Nr > 100: 
            div_i = 50
        elif Nr > 10:
            div_i = 4
        else:
            div_i = 2
        if i%div_i == 0:
            end_inverse_Hankel = time.time()
            time_diff = end_inverse_Hankel - start_inverse_Hankel
            if search_s_max == 'n' and search_Nm == 'n':
                print('   ',"{:3}".format(int(i/Nr*100)),'% done (',"{:.2f}".format(time_diff),'sec)')
            
        for m in range(1,Nm+1):
            sJN[m-1,0] = s[m-1] * mp.besselj( 0, s[m-1] * tr[i-1] ) 
            sJN[m-1,1] = s[m-1] * mp.besselj( 1, s[m-1] * tr[i-1] ) 
        
        for j in range(1,Nz+1):
            
            swt = 0
            for m in range(1,Nm):

                tau_a    [i-1,j-1] += ( sJN[m-1,0] * bta       [m-1,j-1] ) * weight[m-1]
                tau_a_r  [i-1,j-1] += ( sJN[m-1,1] * bta_r     [m-1,j-1] ) * weight[m-1]
                tau_a_z  [i-1,j-1] += ( sJN[m-1,0] * bta_z     [m-1,j-1] ) * weight[m-1]
                tau_b1   [i-1,j-1] += ( sJN[m-1,0] * btb1      [m-1,j-1] ) * weight[m-1]
                tau_b1_r [i-1,j-1] += ( sJN[m-1,1] * btb1_r    [m-1,j-1] ) * weight[m-1]
                tau_b1_z [i-1,j-1] += ( sJN[m-1,0] * btb1_z    [m-1,j-1] ) * weight[m-1]
                tau_b2   [i-1,j-1] += ( sJN[m-1,0] * btb2      [m-1,j-1] ) * weight[m-1]
                tau_b2_r [i-1,j-1] += ( sJN[m-1,1] * btb2_r    [m-1,j-1] ) * weight[m-1]
                tau_b2_z [i-1,j-1] += ( sJN[m-1,0] * btb2_z    [m-1,j-1] ) * weight[m-1]
                u_z_a_c  [i-1,j-1] += ( sJN[m-1,0] * bu_z_a_c  [m-1,j-1] ) * weight[m-1]
                u_r_a_c  [i-1,j-1] += ( sJN[m-1,1] * bu_r_a_c  [m-1,j-1] ) * weight[m-1]
                u_z_a_d1 [i-1,j-1] += ( sJN[m-1,0] * bu_z_a_d1 [m-1,j-1] ) * weight[m-1]
                u_r_a_d1 [i-1,j-1] += ( sJN[m-1,1] * bu_r_a_d1 [m-1,j-1] ) * weight[m-1]
                u_z_a_d2 [i-1,j-1] += ( sJN[m-1,0] * bu_z_a_d2 [m-1,j-1] ) * weight[m-1]
                u_r_a_d2 [i-1,j-1] += ( sJN[m-1,1] * bu_r_a_d2 [m-1,j-1] ) * weight[m-1]
                u_z_b1_c [i-1,j-1] += ( sJN[m-1,0] * bu_z_b1_c [m-1,j-1] ) * weight[m-1]
                u_r_b1_c [i-1,j-1] += ( sJN[m-1,1] * bu_r_b1_c [m-1,j-1] ) * weight[m-1]
                u_z_b1_d1[i-1,j-1] += ( sJN[m-1,0] * bu_z_b1_d1[m-1,j-1] ) * weight[m-1]
                u_r_b1_d1[i-1,j-1] += ( sJN[m-1,1] * bu_r_b1_d1[m-1,j-1] ) * weight[m-1]
                u_z_b1_d2[i-1,j-1] += ( sJN[m-1,0] * bu_z_b1_d2[m-1,j-1] ) * weight[m-1]
                u_r_b1_d2[i-1,j-1] += ( sJN[m-1,1] * bu_r_b1_d2[m-1,j-1] ) * weight[m-1]
                u_z_b2_c [i-1,j-1] += ( sJN[m-1,0] * bu_z_b2_c [m-1,j-1] ) * weight[m-1]
                u_r_b2_c [i-1,j-1] += ( sJN[m-1,1] * bu_r_b2_c [m-1,j-1] ) * weight[m-1]
                u_z_b2_d1[i-1,j-1] += ( sJN[m-1,0] * bu_z_b2_d1[m-1,j-1] ) * weight[m-1]
                u_r_b2_d1[i-1,j-1] += ( sJN[m-1,1] * bu_r_b2_d1[m-1,j-1] ) * weight[m-1]
                u_z_b2_d2[i-1,j-1] += ( sJN[m-1,0] * bu_z_b2_d2[m-1,j-1] ) * weight[m-1]
                u_r_b2_d2[i-1,j-1] += ( sJN[m-1,1] * bu_r_b2_d2[m-1,j-1] ) * weight[m-1]
                
                #chk_??? is the quantity of ??? obtained by using the smaller integration range w.r.t s
                #if |chk_??? - ???| is small, then, it means smaller "s_max0" may be used
                #this check is done by a function "chk_s"
                if swt == 0 and s[m-1] > s_max*mp.mpf('0.5'):
                    swt = 1
                    chk_tau_a    [i-1,j-1] = tau_a    [i-1,j-1]
                    chk_tau_a_r  [i-1,j-1] = tau_a_r  [i-1,j-1]
                    chk_tau_a_z  [i-1,j-1] = tau_a_z  [i-1,j-1]
                    chk_tau_b1   [i-1,j-1] = tau_b1   [i-1,j-1]
                    chk_tau_b1_r [i-1,j-1] = tau_b1_r [i-1,j-1]
                    chk_tau_b1_z [i-1,j-1] = tau_b1_z [i-1,j-1]
                    chk_tau_b2   [i-1,j-1] = tau_b2   [i-1,j-1]
                    chk_tau_b2_r [i-1,j-1] = tau_b2_r [i-1,j-1]
                    chk_tau_b2_z [i-1,j-1] = tau_b2_z [i-1,j-1]
                    chk_u_z_a_c  [i-1,j-1] = u_z_a_c  [i-1,j-1]
                    chk_u_r_a_c  [i-1,j-1] = u_r_a_c  [i-1,j-1]
                    chk_u_z_a_d1 [i-1,j-1] = u_z_a_d1 [i-1,j-1]
                    chk_u_r_a_d1 [i-1,j-1] = u_r_a_d1 [i-1,j-1]
                    chk_u_z_a_d2 [i-1,j-1] = u_z_a_d2 [i-1,j-1]
                    chk_u_r_a_d2 [i-1,j-1] = u_r_a_d2 [i-1,j-1]
                    chk_u_z_b1_c [i-1,j-1] = u_z_b1_c [i-1,j-1]
                    chk_u_r_b1_c [i-1,j-1] = u_r_b1_c [i-1,j-1]
                    chk_u_z_b1_d1[i-1,j-1] = u_z_b1_d1[i-1,j-1]
                    chk_u_r_b1_d1[i-1,j-1] = u_r_b1_d1[i-1,j-1]
                    chk_u_z_b1_d2[i-1,j-1] = u_z_b1_d2[i-1,j-1]
                    chk_u_r_b1_d2[i-1,j-1] = u_r_b1_d2[i-1,j-1]
                    chk_u_z_b2_c [i-1,j-1] = u_z_b2_c [i-1,j-1]
                    chk_u_r_b2_c [i-1,j-1] = u_r_b2_c [i-1,j-1]
                    chk_u_z_b2_d1[i-1,j-1] = u_z_b2_d1[i-1,j-1]
                    chk_u_r_b2_d1[i-1,j-1] = u_r_b2_d1[i-1,j-1]
                    chk_u_z_b2_d2[i-1,j-1] = u_z_b2_d2[i-1,j-1]
                    chk_u_r_b2_d2[i-1,j-1] = u_r_b2_d2[i-1,j-1]

        for j in range(1,Nz_1+1):        
            for m in range(1,Nm):
                tau_a_1 [i-1,j-1] += ( sJN[m-1,0] * bta_1 [m-1,j-1] ) * weight[m-1]
                tau_b1_1[i-1,j-1] += ( sJN[m-1,0] * btb1_1[m-1,j-1] ) * weight[m-1]
                tau_b2_1[i-1,j-1] += ( sJN[m-1,0] * btb2_1[m-1,j-1] ) * weight[m-1]
        
        for j in range(1,Nz_2+1):        
            for m in range(1,Nm):
                tau_a_2 [i-1,j-1] += ( sJN[m-1,0] * bta_2 [m-1,j-1] ) * weight[m-1]
                tau_b1_2[i-1,j-1] += ( sJN[m-1,0] * btb1_2[m-1,j-1] ) * weight[m-1]
                tau_b2_2[i-1,j-1] += ( sJN[m-1,0] * btb2_2[m-1,j-1] ) * weight[m-1]
    
    end_inverse_Hankel = time.time()
    time_diff = end_inverse_Hankel - start_inverse_Hankel
    if search_s_max == 'n' and search_Nm == 'n':
        print('    *** inverse Hankel takes',"{:.2f}".format(time_diff),'sec ***')

    if search_s_max == 'n' and search_Nm == 'n':
        save_macro_nd()
    
    save_macro()
    find_max()
    
def find_max():
    #compute maximum value in (r,z) plane
    global max_tau_a    
    global max_tau_a_r  
    global max_tau_a_z  
    global max_tau_b1   
    global max_tau_b1_r 
    global max_tau_b1_z 
    global max_tau_b2   
    global max_tau_b2_r 
    global max_tau_b2_z 
    global max_u_z_a_c  
    global max_u_r_a_c  
    global max_u_z_a_d1 
    global max_u_r_a_d1 
    global max_u_z_a_d2 
    global max_u_r_a_d2 
    global max_u_z_b1_c 
    global max_u_r_b1_c 
    global max_u_z_b1_d1
    global max_u_r_b1_d1
    global max_u_z_b1_d2
    global max_u_r_b1_d2
    global max_u_z_b2_c 
    global max_u_r_b2_c 
    global max_u_z_b2_d1
    global max_u_r_b2_d1
    global max_u_z_b2_d2
    global max_u_r_b2_d2
    global max_dT
    global max_Tr
    global max_Tz
    global max_vr
    global max_vz
    max_tau_a     = small
    max_tau_a_r   = small
    max_tau_a_z   = small
    max_tau_b1    = small
    max_tau_b1_r  = small
    max_tau_b1_z  = small
    max_tau_b2    = small
    max_tau_b2_r  = small
    max_tau_b2_z  = small
    max_u_z_a_c   = small
    max_u_r_a_c   = small
    max_u_z_a_d1  = small
    max_u_r_a_d1  = small
    max_u_z_a_d2  = small
    max_u_r_a_d2  = small
    max_u_z_b1_c  = small
    max_u_r_b1_c  = small
    max_u_z_b1_d1 = small
    max_u_r_b1_d1 = small
    max_u_z_b1_d2 = small
    max_u_r_b1_d2 = small
    max_u_z_b2_c  = small
    max_u_r_b2_c  = small
    max_u_z_b2_d1 = small
    max_u_r_b2_d1 = small
    max_u_z_b2_d2 = small
    max_u_r_b2_d2 = small
    max_dT        = small
    max_Tr        = small
    max_Tz        = small
    max_vr        = small
    max_vz        = small        
    for i in range(1,Nr+1):
        for j in range(1,Nz+1):
            max_tau_a     = max( max_tau_a    , mp.fabs( tau_a    [i-1,j-1] ) )
            max_tau_a_r   = max( max_tau_a_r  , mp.fabs( tau_a_r  [i-1,j-1] ) )
            max_tau_a_z   = max( max_tau_a_z  , mp.fabs( tau_a_z  [i-1,j-1] ) )
            max_tau_b1    = max( max_tau_b1   , mp.fabs( tau_b1   [i-1,j-1] ) )
            max_tau_b1_r  = max( max_tau_b1_r , mp.fabs( tau_b1_r [i-1,j-1] ) )
            max_tau_b1_z  = max( max_tau_b1_z , mp.fabs( tau_b1_z [i-1,j-1] ) )
            max_tau_b2    = max( max_tau_b2   , mp.fabs( tau_b2   [i-1,j-1] ) )
            max_tau_b2_r  = max( max_tau_b2_r , mp.fabs( tau_b2_r [i-1,j-1] ) )
            max_tau_b2_z  = max( max_tau_b2_z , mp.fabs( tau_b2_z [i-1,j-1] ) )
            max_u_z_a_c   = max( max_u_z_a_c  , mp.fabs( u_z_a_c  [i-1,j-1] ) )
            max_u_r_a_c   = max( max_u_r_a_c  , mp.fabs( u_r_a_c  [i-1,j-1] ) )
            max_u_z_a_d1  = max( max_u_z_a_d1 , mp.fabs( u_z_a_d1 [i-1,j-1] ) )
            max_u_r_a_d1  = max( max_u_r_a_d1 , mp.fabs( u_r_a_d1 [i-1,j-1] ) )
            max_u_z_a_d2  = max( max_u_z_a_d2 , mp.fabs( u_z_a_d2 [i-1,j-1] ) )
            max_u_r_a_d2  = max( max_u_r_a_d2 , mp.fabs( u_r_a_d2 [i-1,j-1] ) )
            max_u_z_b1_c  = max( max_u_z_b1_c , mp.fabs( u_z_b1_c [i-1,j-1] ) )
            max_u_r_b1_c  = max( max_u_r_b1_c , mp.fabs( u_r_b1_c [i-1,j-1] ) )
            max_u_z_b1_d1 = max( max_u_z_b1_d1, mp.fabs( u_z_b1_d1[i-1,j-1] ) )
            max_u_r_b1_d1 = max( max_u_r_b1_d1, mp.fabs( u_r_b1_d1[i-1,j-1] ) )
            max_u_z_b1_d2 = max( max_u_z_b1_d2, mp.fabs( u_z_b1_d2[i-1,j-1] ) )
            max_u_r_b1_d2 = max( max_u_r_b1_d2, mp.fabs( u_r_b1_d2[i-1,j-1] ) )
            max_u_z_b2_c  = max( max_u_z_b2_c , mp.fabs( u_z_b2_c [i-1,j-1] ) )
            max_u_r_b2_c  = max( max_u_r_b2_c , mp.fabs( u_r_b2_c [i-1,j-1] ) )
            max_u_z_b2_d1 = max( max_u_z_b2_d1, mp.fabs( u_z_b2_d1[i-1,j-1] ) )
            max_u_r_b2_d1 = max( max_u_r_b2_d1, mp.fabs( u_r_b2_d1[i-1,j-1] ) )
            max_u_z_b2_d2 = max( max_u_z_b2_d2, mp.fabs( u_z_b2_d2[i-1,j-1] ) )
            max_u_r_b2_d2 = max( max_u_r_b2_d2, mp.fabs( u_r_b2_d2[i-1,j-1] ) )
            max_dT        = max( max_dT       , mp.fabs( dT       [i-1,j-1] ) )
            max_Tr        = max( max_Tr       , mp.fabs( Tr       [i-1,j-1] ) )
            max_Tz        = max( max_Tz       , mp.fabs( Tz       [i-1,j-1] ) )
            max_vr        = max( max_vr       , mp.fabs( vr       [i-1,j-1] ) )
            max_vz        = max( max_vz       , mp.fabs( vz       [i-1,j-1] ) )
    
def chk_s():
    err_max = mp.mpf('0')
    for i in range(1,Nr+1):
        for j in range(1,Nz+1):
            err_max = max( err_max, ( mp.fabs( chk_tau_a    [i-1,j-1] - tau_a    [i-1,j-1] )/max_tau_a     ) )
            err_max = max( err_max, ( mp.fabs( chk_tau_a_r  [i-1,j-1] - tau_a_r  [i-1,j-1] )/max_tau_a_r   ) )
            err_max = max( err_max, ( mp.fabs( chk_tau_a_z  [i-1,j-1] - tau_a_z  [i-1,j-1] )/max_tau_a_z   ) )
            err_max = max( err_max, ( mp.fabs( chk_tau_b1   [i-1,j-1] - tau_b1   [i-1,j-1] )/max_tau_b1    ) )
            err_max = max( err_max, ( mp.fabs( chk_tau_b1_r [i-1,j-1] - tau_b1_r [i-1,j-1] )/max_tau_b1_r  ) )
            err_max = max( err_max, ( mp.fabs( chk_tau_b1_z [i-1,j-1] - tau_b1_z [i-1,j-1] )/max_tau_b1_z  ) )
            err_max = max( err_max, ( mp.fabs( chk_tau_b2   [i-1,j-1] - tau_b2   [i-1,j-1] )/max_tau_b2    ) )
            err_max = max( err_max, ( mp.fabs( chk_tau_b2_r [i-1,j-1] - tau_b2_r [i-1,j-1] )/max_tau_b2_r  ) )
            err_max = max( err_max, ( mp.fabs( chk_tau_b2_z [i-1,j-1] - tau_b2_z [i-1,j-1] )/max_tau_b2_z  ) )
            err_max = max( err_max, ( mp.fabs( chk_u_z_a_c  [i-1,j-1] - u_z_a_c  [i-1,j-1] )/max_u_z_a_c   ) )
            err_max = max( err_max, ( mp.fabs( chk_u_r_a_c  [i-1,j-1] - u_r_a_c  [i-1,j-1] )/max_u_r_a_c   ) )
            err_max = max( err_max, ( mp.fabs( chk_u_z_a_d1 [i-1,j-1] - u_z_a_d1 [i-1,j-1] )/max_u_z_a_d1  ) )
            err_max = max( err_max, ( mp.fabs( chk_u_r_a_d1 [i-1,j-1] - u_r_a_d1 [i-1,j-1] )/max_u_r_a_d1  ) )
            err_max = max( err_max, ( mp.fabs( chk_u_z_a_d2 [i-1,j-1] - u_z_a_d2 [i-1,j-1] )/max_u_z_a_d2  ) )
            err_max = max( err_max, ( mp.fabs( chk_u_r_a_d2 [i-1,j-1] - u_r_a_d2 [i-1,j-1] )/max_u_r_a_d2  ) )
            err_max = max( err_max, ( mp.fabs( chk_u_z_b1_c [i-1,j-1] - u_z_b1_c [i-1,j-1] )/max_u_z_b1_c  ) )
            err_max = max( err_max, ( mp.fabs( chk_u_r_b1_c [i-1,j-1] - u_r_b1_c [i-1,j-1] )/max_u_r_b1_c  ) )
            err_max = max( err_max, ( mp.fabs( chk_u_z_b1_d1[i-1,j-1] - u_z_b1_d1[i-1,j-1] )/max_u_z_b1_d1 ) )
            err_max = max( err_max, ( mp.fabs( chk_u_r_b1_d1[i-1,j-1] - u_r_b1_d1[i-1,j-1] )/max_u_r_b1_d1 ) )
            err_max = max( err_max, ( mp.fabs( chk_u_z_b1_d2[i-1,j-1] - u_z_b1_d2[i-1,j-1] )/max_u_z_b1_d2 ) )
            err_max = max( err_max, ( mp.fabs( chk_u_r_b1_d2[i-1,j-1] - u_r_b1_d2[i-1,j-1] )/max_u_r_b1_d2 ) )
            err_max = max( err_max, ( mp.fabs( chk_u_z_b2_c [i-1,j-1] - u_z_b2_c [i-1,j-1] )/max_u_z_b2_c  ) )
            err_max = max( err_max, ( mp.fabs( chk_u_r_b2_c [i-1,j-1] - u_r_b2_c [i-1,j-1] )/max_u_r_b2_c  ) )
            err_max = max( err_max, ( mp.fabs( chk_u_z_b2_d1[i-1,j-1] - u_z_b2_d1[i-1,j-1] )/max_u_z_b2_d1 ) )
            err_max = max( err_max, ( mp.fabs( chk_u_r_b2_d1[i-1,j-1] - u_r_b2_d1[i-1,j-1] )/max_u_r_b2_d1 ) )
            err_max = max( err_max, ( mp.fabs( chk_u_z_b2_d2[i-1,j-1] - u_z_b2_d2[i-1,j-1] )/max_u_z_b2_d2 ) )
            err_max = max( err_max, ( mp.fabs( chk_u_r_b2_d2[i-1,j-1] - u_r_b2_d2[i-1,j-1] )/max_u_r_b2_d2 ) )
    return err_max

def chk_Nm():
    #err_max = mp.mpf('0')
    err_max = mp.matrix(1,pp+1)# 0 .. pp
    err_max[pp] = mp.mpf('0e0')
    for q in range(pp):# 0, pp-1
        for i in range(1,Nr+1):
            for j in range(1,Nz+1):
                #print(err_max[q], ( mp.fabs( dT_old[i-1,j-1,q] - dT[i-1,j-1] )/max_dT))
                err_max[q] = max( err_max[q], ( mp.fabs( dT_old[i-1,j-1,q] - dT[i-1,j-1] )/max_dT ) )
                err_max[q] = max( err_max[q], ( mp.fabs( Tr_old[i-1,j-1,q] - Tr[i-1,j-1] )/max_Tr ) )
                err_max[q] = max( err_max[q], ( mp.fabs( Tz_old[i-1,j-1,q] - Tz[i-1,j-1] )/max_Tz ) )
                err_max[q] = max( err_max[q], ( mp.fabs( vr_old[i-1,j-1,q] - vr[i-1,j-1] )/max_vr ) )
                err_max[q] = max( err_max[q], ( mp.fabs( vz_old[i-1,j-1,q] - vz[i-1,j-1] )/max_vz ) )

    for i in range(1,Nr+1):
        for j in range(1,Nz+1):
            dT_old[i-1,j-1,pp] = dT[i-1,j-1] 
            Tr_old[i-1,j-1,pp] = Tr[i-1,j-1] 
            Tz_old[i-1,j-1,pp] = Tz[i-1,j-1] 
            vr_old[i-1,j-1,pp] = vr[i-1,j-1] 
            vz_old[i-1,j-1,pp] = vz[i-1,j-1] 
                    
    if pp > 1:
        plt.plot(Nm_array[0:pp],err_max[0:pp],marker='o')
        plt.xscale('log')
        plt.yscale('log')
        plt.xlabel('Nm')
        plt.title('Nm = '+"{:3}".format(Nm))
        plt.ylabel('rel. error between Nm='+"{:3}".format(Nm)+' and lower Nm')
        plt.tight_layout()
        plt.show()
    
    if pp < 2:
        err_Nm = mp.mpf('1e9')
    else:
        err_Nm = err_max[pp-2]
        
    return err_Nm

def save_parameters()-> None:
    if display_macro_val_on_console == 'n':
        print('>> [OUTPUT] non-dimensional paramters -> parameters.dat')
        file = open(fname+"_parameters.dat", "w+", encoding='utf-8', newline='\n')
        file.write('z_R       '); file.write(mp.nstr(z_R       )); file.write('\n');
        file.write('L_0       '); file.write(mp.nstr(L_0       )); file.write('\n');
        file.write('I_0       '); file.write(mp.nstr(I_0       )); file.write('\n');
        file.write('delta_Ta  '); file.write(mp.nstr(delta_Ta  )); file.write('\n');
        file.write('delta_Tb1 '); file.write(mp.nstr(delta_Tb1 )); file.write('\n');
        file.write('delta_Tb2 '); file.write(mp.nstr(delta_Tb2 )); file.write('\n');
        file.write('v_0       '); file.write(mp.nstr(v_0       )); file.write('\n');
        file.write('delta_uc  '); file.write(mp.nstr(delta_uc  )); file.write('\n');
        file.write('delta_ud1 '); file.write(mp.nstr(delta_ud1 )); file.write('\n');
        file.write('delta_ud2 '); file.write(mp.nstr(delta_ud2 )); file.write('\n');
        file.write('Re        '); file.write(mp.nstr(Re        )); file.write('\n');
        file.write('p_0       '); file.write(mp.nstr(p_0       )); file.write('\n');
        file.write('tH        '); file.write(mp.nstr(tH        )); file.write('\n');
        file.write('tH_1      '); file.write(mp.nstr(tH_1      )); file.write('\n');
        file.write('tH_2      '); file.write(mp.nstr(tH_2      )); file.write('\n');
        file.write('tH_m1     '); file.write(mp.nstr(tH_m1     )); file.write('\n');
        file.write('tH_m2     '); file.write(mp.nstr(tH_m2     )); file.write('\n');
        file.write('tk_1      '); file.write(mp.nstr(tk_1      )); file.write('\n');
        file.write('tk_2      '); file.write(mp.nstr(tk_2      )); file.write('\n');
        file.write('tk_m1     '); file.write(mp.nstr(tk_m1     )); file.write('\n');
        file.write('tk_m2     '); file.write(mp.nstr(tk_m2     )); file.write('\n');
        file.write('tz_R      '); file.write(mp.nstr(tz_R      )); file.write('\n');
        file.write('tz_R2     '); file.write(mp.nstr(tz_R2     )); file.write('\n');
        file.write('tz_R4     '); file.write(mp.nstr(tz_R4     )); file.write('\n');
        file.write('tz_0      '); file.write(mp.nstr(tz_0      )); file.write('\n');
        file.write('tD_T      '); file.write(mp.nstr(tD_T      )); file.write('\n');
        file.write('F_0       '); file.write(mp.nstr(F_0       )); file.write('\n');
        file.write('s_max     '); file.write(mp.nstr(s_max     )); file.write('\n');
        file.write('recommended s_max0 '); file.write(mp.nstr(s_max/(z_R/H))); file.write('\n');
        file.write('recommended Nm     '); file.write(mp.nstr(Nm           )); file.write('\n');
        file.write('cpu-time           '); file.write("{:.1f}".format(time_diff)+' sec');
        file.close()
    else:
        print('>> [CHECK] non-dimensional paramters')
        print('   z_R       = '+mp.nstr(z_R       ))
        print('   L_0       = '+mp.nstr(L_0       ))
        print('   I_0       = '+mp.nstr(I_0       ))
        print('   delta_Ta  = '+mp.nstr(delta_Ta  ))
        print('   delta_Tb1 = '+mp.nstr(delta_Tb1 ))
        print('   delta_Tb2 = '+mp.nstr(delta_Tb2 ))
        print('   v_0       = '+mp.nstr(v_0       ))
        print('   delta_uc  = '+mp.nstr(delta_uc  ))
        print('   delta_ud1 = '+mp.nstr(delta_ud1 ))
        print('   delta_ud2 = '+mp.nstr(delta_ud2 ))
        print('   Re        = '+mp.nstr(Re        ))
        print('   p_0       = '+mp.nstr(p_0       ))
        print('   tH        = '+mp.nstr(tH        ))
        print('   tH_1      = '+mp.nstr(tH_1      ))
        print('   tH_2      = '+mp.nstr(tH_2      ))
        print('   tH_m1     = '+mp.nstr(tH_m1     ))
        print('   tH_m2     = '+mp.nstr(tH_m2     ))
        print('   tk_1      = '+mp.nstr(tk_1      ))
        print('   tk_2      = '+mp.nstr(tk_2      ))
        print('   tk_m1     = '+mp.nstr(tk_m1     ))
        print('   tk_m2     = '+mp.nstr(tk_m2     ))
        print('   tz_R      = '+mp.nstr(tz_R      ))
        print('   tz_R2     = '+mp.nstr(tz_R2     ))
        print('   tz_R4     = '+mp.nstr(tz_R4     ))
        print('   tz_0      = '+mp.nstr(tz_0      ))
        print('   tD_T      = '+mp.nstr(tD_T      ))
        print('   F_0       = '+mp.nstr(F_0       ))
        print('   s_max     = '+mp.nstr(s_max     ))
        print('   recommended s_max0 = '+mp.nstr(s_max/(z_R/H)))
        print('   recommended Nm     = '+mp.nstr(Nm           ))
    
def save_macro_nd()-> None:
    if display_macro_val_on_console == 'n':
        print('>> [OUTPUT] non-dimensional quantities -> macro_nd.dat')              
        file = open(fname+"_macro_nd.dat", "w+", encoding='utf-8', newline='\n')
        file.write('tau_a    \n'); file.write(mp.nstr(delta_Ta *tau_a              )); file.write('\n');
        file.write('tau_a_r  \n'); file.write(mp.nstr(delta_Ta *tau_a_r            )); file.write('\n');
        file.write('tau_a_z  \n'); file.write(mp.nstr(delta_Ta *tau_a_z            )); file.write('\n');
        file.write('tau_b1   \n'); file.write(mp.nstr(delta_Tb1*tau_b1             )); file.write('\n');
        file.write('tau_b1_r \n'); file.write(mp.nstr(delta_Tb1*tau_b1_r           )); file.write('\n');
        file.write('tau_b1_z \n'); file.write(mp.nstr(delta_Tb1*tau_b1_z           )); file.write('\n');
        file.write('tau_b2   \n'); file.write(mp.nstr(delta_Tb2*tau_b2             )); file.write('\n');
        file.write('tau_b2_r \n'); file.write(mp.nstr(delta_Tb2*tau_b2_r           )); file.write('\n');
        file.write('tau_b2_z \n'); file.write(mp.nstr(delta_Tb2*tau_b2_z           )); file.write('\n');
        file.write('tau_a_1  \n'); file.write(mp.nstr(delta_Ta *tau_a_1            )); file.write('\n');
        file.write('tau_b1_1 \n'); file.write(mp.nstr(delta_Tb1*tau_b1_1           )); file.write('\n');
        file.write('tau_b2_1 \n'); file.write(mp.nstr(delta_Tb2*tau_b2_1           )); file.write('\n');
        file.write('tau_a_2  \n'); file.write(mp.nstr(delta_Ta *tau_a_2            )); file.write('\n');
        file.write('tau_b1_2 \n'); file.write(mp.nstr(delta_Tb1*tau_b1_2           )); file.write('\n');
        file.write('tau_b2_2 \n'); file.write(mp.nstr(delta_Tb2*tau_b2_2           )); file.write('\n');
        file.write('u_z_a_c  \n'); file.write(mp.nstr(delta_Ta *delta_uc *u_z_a_c  )); file.write('\n');
        file.write('u_r_a_c  \n'); file.write(mp.nstr(delta_Ta *delta_uc *u_r_a_c  )); file.write('\n');
        file.write('u_z_a_d1 \n'); file.write(mp.nstr(delta_Ta *delta_ud1*u_z_a_d1 )); file.write('\n');
        file.write('u_r_a_d1 \n'); file.write(mp.nstr(delta_Ta *delta_ud1*u_r_a_d1 )); file.write('\n');
        file.write('u_z_a_d2 \n'); file.write(mp.nstr(delta_Ta *delta_ud2*u_z_a_d2 )); file.write('\n');
        file.write('u_r_a_d2 \n'); file.write(mp.nstr(delta_Ta *delta_ud2*u_r_a_d2 )); file.write('\n');
        file.write('u_z_b1_c \n'); file.write(mp.nstr(delta_Tb1*delta_uc *u_z_b1_c )); file.write('\n');
        file.write('u_r_b1_c \n'); file.write(mp.nstr(delta_Tb1*delta_uc *u_r_b1_c )); file.write('\n');
        file.write('u_z_b1_d1\n'); file.write(mp.nstr(delta_Tb1*delta_ud1*u_z_b1_d1)); file.write('\n');
        file.write('u_r_b1_d1\n'); file.write(mp.nstr(delta_Tb1*delta_ud1*u_r_b1_d1)); file.write('\n');
        file.write('u_z_b1_d2\n'); file.write(mp.nstr(delta_Tb1*delta_ud2*u_z_b1_d2)); file.write('\n');
        file.write('u_r_b1_d2\n'); file.write(mp.nstr(delta_Tb1*delta_ud2*u_r_b1_d2)); file.write('\n');
        file.write('u_z_b2_c \n'); file.write(mp.nstr(delta_Tb2*delta_uc *u_z_b2_c )); file.write('\n');
        file.write('u_r_b2_c \n'); file.write(mp.nstr(delta_Tb2*delta_uc *u_r_b2_c )); file.write('\n');
        file.write('u_z_b2_d1\n'); file.write(mp.nstr(delta_Tb2*delta_ud1*u_z_b2_d1)); file.write('\n');
        file.write('u_r_b2_d1\n'); file.write(mp.nstr(delta_Tb2*delta_ud1*u_r_b2_d1)); file.write('\n');
        file.write('u_z_b2_d2\n'); file.write(mp.nstr(delta_Tb2*delta_ud2*u_z_b2_d2)); file.write('\n');
        file.write('u_r_b2_d2\n'); file.write(mp.nstr(delta_Tb2*delta_ud2*u_r_b2_d2)); file.write('\n');
        file.close()
    
def save_macro()-> None:

    for i in range(1,Nr+1):
        for j in range(1,Nz+1):
            dT[i-1,j-1] =   T_0       * delta_Ta  * tau_a [i-1,j-1] \
                          + T_0       * delta_Tb1 * tau_b1[i-1,j-1] \
                          + T_0       * delta_Tb2 * tau_b2[i-1,j-1] 
            Tr[i-1,j-1] =   T_0 / w_0 * delta_Ta  * tau_a_r [i-1,j-1] \
                          + T_0 / w_0 * delta_Tb1 * tau_b1_r[i-1,j-1] \
                          + T_0 / w_0 * delta_Tb2 * tau_b2_r[i-1,j-1] 
            Tz[i-1,j-1] =   T_0 / w_0 * delta_Ta  * tau_a_z [i-1,j-1] \
                          + T_0 / w_0 * delta_Tb1 * tau_b1_z[i-1,j-1] \
                          + T_0 / w_0 * delta_Tb2 * tau_b2_z[i-1,j-1] 
            vr[i-1,j-1] =   v_0       * delta_Ta  * delta_uc  * u_r_a_c  [i-1,j-1] \
                          + v_0       * delta_Tb1 * delta_uc  * u_r_b1_c [i-1,j-1] \
                          + v_0       * delta_Tb2 * delta_uc  * u_r_b2_c [i-1,j-1] \
                          + v_0       * delta_Ta  * delta_ud1 * u_r_a_d1 [i-1,j-1] \
                          + v_0       * delta_Tb1 * delta_ud1 * u_r_b1_d1[i-1,j-1] \
                          + v_0       * delta_Tb2 * delta_ud1 * u_r_b2_d1[i-1,j-1] \
                          + v_0       * delta_Ta  * delta_ud2 * u_r_a_d2 [i-1,j-1] \
                          + v_0       * delta_Tb1 * delta_ud2 * u_r_b1_d2[i-1,j-1] \
                          + v_0       * delta_Tb2 * delta_ud2 * u_r_b2_d2[i-1,j-1] 
            vz[i-1,j-1] =   v_0       * delta_Ta  * delta_uc  * u_z_a_c  [i-1,j-1] \
                          + v_0       * delta_Tb1 * delta_uc  * u_z_b1_c [i-1,j-1] \
                          + v_0       * delta_Tb2 * delta_uc  * u_z_b2_c [i-1,j-1] \
                          + v_0       * delta_Ta  * delta_ud1 * u_z_a_d1 [i-1,j-1] \
                          + v_0       * delta_Tb1 * delta_ud1 * u_z_b1_d1[i-1,j-1] \
                          + v_0       * delta_Tb2 * delta_ud1 * u_z_b2_d1[i-1,j-1] \
                          + v_0       * delta_Ta  * delta_ud2 * u_z_a_d2 [i-1,j-1] \
                          + v_0       * delta_Tb1 * delta_ud2 * u_z_b1_d2[i-1,j-1] \
                          + v_0       * delta_Tb2 * delta_ud2 * u_z_b2_d2[i-1,j-1] 
                          
    for i in range(1,Nr+1):
        for j in range(1,Nz_1+1):
            dT_1[i-1,j-1] =   T_0 * delta_Ta  * tau_a_1 [i-1,j-1] \
                            + T_0 * delta_Tb1 * tau_b1_1[i-1,j-1] \
                            + T_0 * delta_Tb2 * tau_b2_1[i-1,j-1] 

    for i in range(1,Nr+1):
        for j in range(1,Nz_2+1):
            dT_2[i-1,j-1] =   T_0 * delta_Ta  * tau_a_2 [i-1,j-1] \
                            + T_0 * delta_Tb1 * tau_b1_2[i-1,j-1] \
                            + T_0 * delta_Tb2 * tau_b2_2[i-1,j-1] 

    if search_s_max == 'n' and search_Nm == 'n':
        #pd.describe_option()
        pd.set_option('display.width', 10000)
        pd.set_option('display.max_rows', None)
        pd.set_option('display.max_columns', None)
        r_float = np.zeros(Nr)
        for i in range(1,Nr+1):
            r_float[i-1] = float(w_0*tr[i-1])
        z_float = np.zeros(Nz)
        for j in range(1,Nz+1):
            z_float[j-1] = float(w_0*tz[j-1])
        z_1_float = np.zeros(Nz_1)
        for j in range(1,Nz_1+1):
            z_1_float[j-1] = float(w_0*tz_1[j-1])
        z_2_float = np.zeros(Nz_2)
        for j in range(1,Nz_2+1):
            z_2_float[j-1] = float(w_0*tz_2[j-1])
        ####    
        df = pd.DataFrame(dT.T,columns=r_float,index=z_float)
        header = "dT(=T-T0)[K] / r[m] (col) and z[m] (row)"
        if display_macro_val_on_console == 'y':
            print(">> "+header)
            print(df.map('{:.6e}'.format))
        else:
            fname_out = fname+"_macro-dT.csv"
            print('>> [OUTPUT] '+header+' > '+fname_out)              
            df.to_csv(fname_out,float_format='%11.6e')
        ####    
        df = pd.DataFrame(Tr.T,columns=r_float,index=z_float)
        header = "Tr(=dT/dr)[K/m] / r[m] (col) and z[m] (row)"
        if display_macro_val_on_console == 'y':
            print(">> "+header)
            print(df.map('{:.6e}'.format))
        else:
            fname_out = fname+"_macro-Tr.csv"
            print('>> [OUTPUT] '+header+' > '+fname_out)              
            df.to_csv(fname_out,float_format='%11.6e')
        ####    
        df = pd.DataFrame(Tz.T,columns=r_float,index=z_float)
        header = "Tz(=dT/dz)[K/m] / r[m] (col) and z[m] (row)"
        if display_macro_val_on_console == 'y':
            print(">> "+header)
            print(df.map('{:.6e}'.format))
        else:
            fname_out = fname+"_macro-Tz.csv"
            print('>> [OUTPUT] '+header+' > '+fname_out)              
            df.to_csv(fname_out,float_format='%11.6e')
        ####    
        df = pd.DataFrame(vr.T,columns=r_float,index=z_float)
        header = "vr[m/s] / r[m] (col) and z[m] (row)"
        if display_macro_val_on_console == 'y':
            print(">> "+header)
            print(df.map('{:.6e}'.format))
        else:
            fname_out = fname+"_macro-vr.csv"
            print('>> [OUTPUT] '+header+' > '+fname_out)              
            df.to_csv(fname_out,float_format='%11.6e')
        ####    
        df = pd.DataFrame(vz.T,columns=r_float,index=z_float)
        header = "vz[m/s] / r[m] (col) and z[m] (row)"
        if display_macro_val_on_console == 'y':
            print(">> "+header)
            print(df.map('{:.6e}'.format))
        else:
            fname_out = fname+"_macro-vz.csv"
            print('>> [OUTPUT] '+header+' > '+fname_out)              
            df.to_csv(fname_out,float_format='%11.6e')
        ####    
        df = pd.DataFrame(dT_1.T,columns=r_float,index=z_1_float)
        header = "dT_1(=T_1-T0)[K] / r[m] (col) and z[m] (row)"
        if display_macro_val_on_console == 'y':
            print(">> "+header)
            print(df.map('{:.6e}'.format))
        else:
            fname_out = fname+"_macro-dT_1.csv"
            print('>> [OUTPUT] '+header+' > '+fname_out)              
            df.to_csv(fname_out,float_format='%11.6e')
        ####    
        df = pd.DataFrame(dT_2.T,columns=r_float,index=z_2_float)
        header = "dT_2(=T_2-T0)[K] / r[m] (col) and z[m] (row)"
        if display_macro_val_on_console == 'y':
            print(">> "+header)
            print(df.map('{:.6e}'.format))
        else:
            fname_out = fname+"_macro-dT_2.csv"
            print('>> [OUTPUT] '+header+' > '+fname_out)              
            df.to_csv(fname_out,float_format='%11.6e')

        # file = open(fname+"_macro.dat", "w+", encoding='utf-8', newline='\n')
        # file.write('dT  \n'); file.write(mp.nstr(dT  )); file.write('\n');
        # file.write('Tr  \n'); file.write(mp.nstr(Tr  )); file.write('\n');
        # file.write('Tz  \n'); file.write(mp.nstr(Tz  )); file.write('\n');
        # file.write('vr  \n'); file.write(mp.nstr(vr  )); file.write('\n');
        # file.write('vz  \n'); file.write(mp.nstr(vz  )); file.write('\n');
        # file.write('dT_1\n'); file.write(mp.nstr(dT_1)); file.write('\n');
        # file.write('dT_2\n'); file.write(mp.nstr(dT_2)); file.write('\n');
        # file.close()

    
def get_C( arg_m, arg_s ):
    sH   = arg_s * tH
    csH  = mp.cosh( sH );  ssH  = mp.sinh( sH ) 
    esHp = mp.exp(  sH );  esHm = mp.exp( -sH )
    F0p = func_F( arg_s,     - tz_0 ); F0m = func_F( -arg_s,     - tz_0 ) 
    FHp = func_F( arg_s, tH  - tz_0 ); FHm = func_F( -arg_s, tH  - tz_0 ) 
    S0p = func_S( arg_s,     - tz_0 ); S0m = func_S( -arg_s,     - tz_0 ) 
    SHp = func_S( arg_s, tH  - tz_0 ); SHm = func_S( -arg_s, tH  - tz_0 ) 
    V0p = func_V( arg_s,     - tz_0 ); V0m = func_V( -arg_s,     - tz_0 ) 
    VHp = func_V( arg_s, tH  - tz_0 ); VHm = func_V( -arg_s, tH  - tz_0 ) 

    res_Yc  = mp.matrix(6,3) 
    res_Yd1 = mp.matrix(6,3) 
    res_Yd2 = mp.matrix(6,3) 

    for ab in range(3):
        for cd in range(3):
            if cd == 0: #C1p and C1m
                res_Yc[0,ab] = ( - sH * sH / mp.mpf('4') ) * ( coef_Bp[arg_m,ab] )
                res_Yc[3,ab] = ( - sH * sH / mp.mpf('4') ) * ( coef_Bm[arg_m,ab] )
                C1p = res_Yc[0,ab]
                C1m = res_Yc[3,ab]
            elif cd == 1: 
                res_Yd1[0,ab] = mp.mpf('0')
                res_Yd1[3,ab] = mp.mpf('0')
                C1p = res_Yd1[0,ab]
                C1m = res_Yd1[3,ab]
            elif cd == 2: 
                res_Yd2[0,ab] = mp.mpf('0')
                res_Yd2[3,ab] = mp.mpf('0')
                C1p = res_Yd2[0,ab]
                C1m = res_Yd2[3,ab]

            if ab == 0 and cd == 0:
                J0  =   F0m * S0m + F0p*S0p
                JH  =   FHm * SHm + FHp*SHp
                dJ0 = ( F0m * V0m + F0p*V0p ) * tH 
                dJH = ( FHm * VHm + FHp*VHp ) * tH 
            elif cd == 1:
                J0  = mp.mpf('0')
                JH  = mp.mpf('0')
                dJ0 = sH
                dJH = mp.mpf('0')
            elif cd == 2:
                J0  = mp.mpf('0')
                JH  = mp.mpf('0')
                dJ0 = mp.mpf('0')
                dJH = sH
            else:
                J0  = mp.mpf('0')
                JH  = mp.mpf('0')
                dJ0 = mp.mpf('0')
                dJH = mp.mpf('0')
                
            Qu = mp.cosh( mp.mpf('2') * sH ) - ( mp.mpf('2') * sH * sH + mp.mpf('1') ) 
    
            #C2p
            P1 =  (     J0 - esHp*JH + dJ0       + F0p*C1p + F0m*C1m ) * ssH 
            P2 =  ( csH*J0 -      JH       + dJH - FHp*C1p - FHm*C1m )       
            P3 =  (               JH                                 )       
            res_Y2 = P3 * sH * sH + P2 * sH + P1
            res_Y2 = res_Y2 * FHm / Qu

            #C2m
            P1 = -(     J0 - esHm*JH + dJ0       + F0p*C1p + F0m*C1m ) * ssH 
            P2 = -( csH*J0 -      JH       + dJH - FHp*C1p - FHm*C1m )       
            P3 =  (               JH                                 )       
            res_Y5 = P3 * sH * sH + P2 * sH + P1
            res_Y5 = res_Y5 * FHp / Qu

            #C3p
            P1 =  ( esHm*J0 -     JH       + dJH - FHp*C1p - FHm*C1m ) * ssH 
            P2 =  (      J0 - csH*JH + dJ0       + F0p*C1p + F0m*C1m )       
            P3 =  (      J0                                          )       
            res_Y3 = P3 * sH * sH + P2 * sH + P1
            res_Y3 = res_Y3 * F0m / Qu

            #C3m
            P1 = -( esHp*J0 -     JH       + dJH - FHp*C1p - FHm*C1m ) * ssH 
            P2 = -(      J0 - csH*JH + dJ0       + F0p*C1p + F0m*C1m )       
            P3 =  (      J0                                          )       
            res_Y6 = P3 * sH * sH + P2 * sH + P1
            res_Y6 = res_Y6 * F0p / Qu 
            
            if cd == 0:
                res_Yc[1,ab] = res_Y2
                res_Yc[4,ab] = res_Y5
                res_Yc[2,ab] = res_Y3
                res_Yc[5,ab] = res_Y6
            elif cd == 1:
                res_Yd1[1,ab] = res_Y2
                res_Yd1[4,ab] = res_Y5
                res_Yd1[2,ab] = res_Y3
                res_Yd1[5,ab] = res_Y6
            elif cd == 2: 
                res_Yd2[1,ab] = res_Y2
                res_Yd2[4,ab] = res_Y5
                res_Yd2[2,ab] = res_Y3
                res_Yd2[5,ab] = res_Y6
    
    return res_Yc, res_Yd1, res_Yd2
    
def get_B( arg_s ):
    mH = max( tH_1, tH_2 )
    GF0p = func_G_F (  arg_s,     - tz_0 )
    GF0m = func_G_F ( -arg_s,     - tz_0 )
    GFHp = func_G_F (  arg_s, tH  - tz_0 )
    GFHm = func_G_F ( -arg_s, tH  - tz_0 )
    EF0p = func_E_F (  arg_s,     - tz_0 )
    EF0m = func_E_F ( -arg_s,     - tz_0 )
    EFHp = func_E_F (  arg_s, tH  - tz_0 )
    EFHm = func_E_F ( -arg_s, tH  - tz_0 )
    auxp = func_F   (  arg_s, tH_1- mH   )# exp( -s*mH) をあらかじめかけておいて，exp が大きくなりすぎるのを押さえている flag mH で回収
    auxm = func_F   ( -arg_s, tH_1+ mH   )
    a0 = auxp - auxm
    b0 = auxp + auxm
    auxp = func_F   (  arg_s, -tH_2- mH )
    auxm = func_F   ( -arg_s, -tH_2+ mH )
    aH = auxp - auxm
    bH = auxp + auxm
    ###
    I0  =   EF0m + EF0p
    IH  =   EFHm + EFHp
    dI0 = - EF0m + EF0p + GF0m - GF0p
    dIH = - EFHm + EFHp + GFHm - GFHp 
    ###
    Hsf1p = mp.mpf('1') + tH_m1 * tH_m1 * mp.mpf('0.25') * arg_s * arg_s
    Hsf2p = mp.mpf('1') + tH_m2 * tH_m2 * mp.mpf('0.25') * arg_s * arg_s
    Hsf1m = mp.mpf('1') - tH_m1 * tH_m1 * mp.mpf('0.25') * arg_s * arg_s
    Hsf2m = mp.mpf('1') - tH_m2 * tH_m2 * mp.mpf('0.25') * arg_s * arg_s
    Hsf1u = ( tH_m1 * tk_m1 ) * arg_s
    Hsf1l = ( tH_m1 / tk_m1 ) * arg_s
    Hsf2u = ( tH_m2 * tk_m2 ) * arg_s
    Hsf2l = ( tH_m2 / tk_m2 ) * arg_s
    ###
    Q1  = - ( Hsf1u * a0 + Hsf1p * ( b0*tk_1 + a0 ) + b0 * tk_1 * Hsf1l ) \
          * ( Hsf2u * aH + Hsf2p * (-bH*tk_2 + aH ) - bH * tk_2 * Hsf2l ) \
          * mp.exp(  arg_s * tH )
    Q2  =   ( Hsf1u * a0 - Hsf1p * (-b0*tk_1 + a0 ) - b0 * tk_1 * Hsf1l ) \
          * ( Hsf2u * aH - Hsf2p * ( bH*tk_2 + aH ) + bH * tk_2 * Hsf2l ) \
          * mp.exp( -arg_s * tH )
    Q   = Q1 + Q2
    ### 
    res_X = mp.matrix(4,3) 
    ### Bm
    EA = mp.exp( (    - tz_0 ) * arg_s )
    EB = mp.exp( ( tH - tz_0 ) * arg_s )
    #*** a
    P1 = -( Hsf1p - Hsf1l )*( Hsf2p * dIH + Hsf2u * IH )*EA + ( -Hsf1l * dI0 + Hsf1p * I0 )*( Hsf2p + Hsf2u )*EB
    P2 = -( Hsf1p - Hsf1u )*( Hsf2l * dIH + Hsf2p * IH )*EA + (  Hsf1p * dI0 - Hsf1u * I0 )*( Hsf2p + Hsf2l )*EB
    P3 =  ( Hsf1p - Hsf1l )*( Hsf2l * dIH + Hsf2p * IH )*EA - ( -Hsf1l * dI0 + Hsf1p * I0 )*( Hsf2p + Hsf2l )*EB
    P4 =  ( Hsf1p - Hsf1u )*( Hsf2p * dIH + Hsf2u * IH )*EA - (  Hsf1p * dI0 - Hsf1u * I0 )*( Hsf2p + Hsf2u )*EB
    res_X[0,0] = (   P1      * aH * tk_1 * b0             \
                   + P2 * a0                  * tk_2 * bH \
                   + P3 *           tk_1 * b0 * tk_2 * bH \
                   + P4 * a0 * aH                         ) / Q
    #*** b1
    P1 =    ( -Hsf1l       )*( Hsf2p + Hsf2u )*EB
    P2 =    (  mp.mpf('2') )*( Hsf2p + Hsf2l )*EB
    P3 =  - ( -Hsf1l     )  *( Hsf2p + Hsf2l )*EB
    P4 =  - (  mp.mpf('2') )*( Hsf2p + Hsf2u )*EB
    res_X[0,1] = (   P1      * aH * tk_1 * b0             \
                   + P2 * a0                  * tk_2 * bH \
                   + P3 *           tk_1 * b0 * tk_2 * bH \
                   + P4 * a0 * aH                         ) / Q
    #*** b2
    P1 = -( Hsf1p - Hsf1l )*( -mp.mpf('2') )*EA 
    P2 = -( Hsf1p - Hsf1u )*( -Hsf2l       )*EA 
    P3 =  ( Hsf1p - Hsf1l )*( -Hsf2l       )*EA 
    P4 =  ( Hsf1p - Hsf1u )*( -mp.mpf('2') )*EA 
    res_X[0,2] = (   P1      * aH * tk_1 * b0             \
                   + P2 * a0                  * tk_2 * bH \
                   + P3 *           tk_1 * b0 * tk_2 * bH \
                   + P4 * a0 * aH                         ) / Q
    
    ### Bp 
    EA = mp.exp( - (    - tz_0 ) * arg_s )
    EB = mp.exp( - ( tH - tz_0 ) * arg_s )
    #*** a
    P1 =  ( Hsf1p + Hsf1l )*( Hsf2p * dIH + Hsf2u * IH )*EA + ( -Hsf1l * dI0 + Hsf1p * I0 )*( Hsf2p - Hsf2u )*EB
    P2 = -( Hsf1p + Hsf1u )*( Hsf2l * dIH + Hsf2p * IH )*EA - (  Hsf1p * dI0 - Hsf1u * I0 )*( Hsf2p - Hsf2l )*EB
    P3 = -( Hsf1p + Hsf1l )*( Hsf2l * dIH + Hsf2p * IH )*EA + ( -Hsf1l * dI0 + Hsf1p * I0 )*( Hsf2p - Hsf2l )*EB
    P4 =  ( Hsf1p + Hsf1u )*( Hsf2p * dIH + Hsf2u * IH )*EA - (  Hsf1p * dI0 - Hsf1u * I0 )*( Hsf2p - Hsf2u )*EB
    res_X[1,0] = (   P1      * aH * tk_1 * b0             \
                   + P2 * a0                  * tk_2 * bH \
                   + P3 *           tk_1 * b0 * tk_2 * bH \
                   + P4 * a0 * aH                         ) / Q
    #*** b1
    P1 =   ( -Hsf1l       )*( Hsf2p - Hsf2u )*EB
    P2 = - (  mp.mpf('2') )*( Hsf2p - Hsf2l )*EB
    P3 =   ( -Hsf1l       )*( Hsf2p - Hsf2l )*EB
    P4 = - (  mp.mpf('2') )*( Hsf2p - Hsf2u )*EB
    res_X[1,1] = (   P1      * aH * tk_1 * b0             \
                   + P2 * a0                  * tk_2 * bH \
                   + P3 *           tk_1 * b0 * tk_2 * bH \
                   + P4 * a0 * aH                         ) / Q
    #*** b2
    P1 =  ( Hsf1p + Hsf1l )*( -mp.mpf('2') )*EA 
    P2 = -( Hsf1p + Hsf1u )*( -Hsf2l       )*EA 
    P3 = -( Hsf1p + Hsf1l )*( -Hsf2l       )*EA 
    P4 =  ( Hsf1p + Hsf1u )*( -mp.mpf('2') )*EA 
    res_X[1,2] = (   P1      * aH * tk_1 * b0             \
                   + P2 * a0                  * tk_2 * bH \
                   + P3 *           tk_1 * b0 * tk_2 * bH \
                   + P4 * a0 * aH                         ) / Q    
    ### B1 
    EA = mp.exp( - arg_s * tH )
    EB = mp.exp(   arg_s * tH )
    #*** a
    P1 = ( ( Hsf2p - Hsf2l )*( I0 - dI0 )*EA + ( dI0 + I0 )*( Hsf2p + Hsf2l )*EB - mp.mpf('2')*( Hsf2l * dIH + Hsf2p * IH ) )*Hsf1m
    P2 = ( ( Hsf2p - Hsf2u )*( I0 - dI0 )*EA - ( dI0 + I0 )*( Hsf2p + Hsf2u )*EB + mp.mpf('2')*( Hsf2p * dIH + Hsf2u * IH ) )*Hsf1m
    res_X[2,0] = ( P1*tk_2*bH + P2*aH ) / ( Q ) 
    #
    sgn        = mp.sign( P1*tk_2*bH + P2*aH ) 
    pow        = mp.log( sgn*( P1*tk_2*bH + P2*aH ) ) - mp.log( Q )
    pow        = pow - ( arg_s * mH ) #flag mH
    res_X[2,0] = mp.exp( pow ) * sgn
    #*** b1
    P1 = ( Hsf2p - Hsf2l )*( Hsf1l - mp.mpf('2') )*EA + ( Hsf1l + mp.mpf('2') )*( Hsf2p + Hsf2l )*EB 
    P2 = ( Hsf2p - Hsf2u )*( Hsf1l - mp.mpf('2') )*EA - ( Hsf1l + mp.mpf('2') )*( Hsf2p + Hsf2u )*EB 
    res_X[2,1] = ( P1*tk_2*bH + P2*aH ) / ( Q ) 
    sgn        = mp.sign( P1*tk_2*bH + P2*aH ) 
    pow        = mp.log( sgn*( P1*tk_2*bH + P2*aH ) ) - mp.log( Q )
    pow        = pow - ( arg_s * mH ) #flag mH
    res_X[2,1] = mp.exp( pow ) * sgn
    #*** b2
    P1 = ( - mp.mpf('2')*( -Hsf2l         ) )*Hsf1m
    P2 = ( + mp.mpf('2')*( -mp.mpf('2')   ) )*Hsf1m
    res_X[2,2] = ( P1*tk_2*bH + P2*aH ) / ( Q ) 
    sgn        = mp.sign( P1*tk_2*bH + P2*aH ) 
    pow        = mp.log( sgn*( P1*tk_2*bH + P2*aH ) ) - mp.log( Q )
    pow        = pow - ( arg_s * mH ) #flag mH
    res_X[2,2] = mp.exp( pow ) * sgn    
    ### B2 
    EA = mp.exp( - arg_s * tH )
    EB = mp.exp(   arg_s * tH )
    #*** a
    P1 = (-( Hsf1p - Hsf1l )*( dIH + IH )*EA - ( IH - dIH )*( Hsf1p + Hsf1l )*EB + mp.mpf('2')*(-Hsf1l * dI0 + Hsf1p * I0 ) )*Hsf2m
    P2 = ( ( Hsf1p - Hsf1u )*( dIH + IH )*EA - ( IH - dIH )*( Hsf1p + Hsf1u )*EB - mp.mpf('2')*( Hsf1p * dI0 - Hsf1u * I0 ) )*Hsf2m
    res_X[3,0] = ( P1*tk_1*b0 + P2*a0 ) / ( Q ) 
    sgn        = mp.sign( P1*tk_1*b0 + P2*a0 )
    pow        = mp.log( sgn*( P1*tk_1*b0 + P2*a0 ) ) - mp.log( Q )
    pow        = pow - ( arg_s * mH ) #flag mH
    res_X[3,0] = mp.exp( pow ) * sgn
    #*** b1
    P1 = (   mp.mpf('2')*( -Hsf1l       ) )*Hsf2m
    P2 = ( - mp.mpf('2')*(  mp.mpf('2') ) )*Hsf2m
    res_X[3,1] = ( P1*tk_1*b0 + P2*a0 ) / ( Q ) 
    sgn        = mp.sign( P1*tk_1*b0 + P2*a0 )
    pow        = mp.log( sgn*( P1*tk_1*b0 + P2*a0 ) ) - mp.log( Q )
    pow        = pow - ( arg_s * mH ) #flag mH
    res_X[3,1] = mp.exp( pow ) * sgn
    #*** b2
    P1 = -( Hsf1p - Hsf1l )*( Hsf2l - mp.mpf('2') )*EA - ( Hsf2l + mp.mpf('2') )*( Hsf1p + Hsf1l )*EB 
    P2 =  ( Hsf1p - Hsf1u )*( Hsf2l - mp.mpf('2') )*EA - ( Hsf2l + mp.mpf('2') )*( Hsf1p + Hsf1u )*EB 
    res_X[3,2] = ( P1*tk_1*b0 + P2*a0 ) / ( Q ) 
    sgn        = mp.sign( P1*tk_1*b0 + P2*a0 )
    pow        = mp.log( sgn*( P1*tk_1*b0 + P2*a0 ) ) - mp.log( Q )
    pow        = pow - ( arg_s * mH ) #flag mH
    res_X[3,2] = mp.exp( pow ) * sgn    
    return res_X    
    
def compute_tausa(arg):
    arg2    = arg * arg
    res     = spi * mp.exp( tz_R2 - arg2 / mp.mpf('8') ) / arg2 / mp.mpf('8') 
    return res

def func_w(arg_z):
    arg = mp.mpf('2') * ( arg_z - tz_0 ) * ( arg_z - tz_0 ) / tz_R / tz_R
    res = mp.sqrt( mp.mpf('1') + arg )
    return res

def compute_tausb1(arg):
    res_w = func_w( mp.mpf('0') )
    arg2    = arg * arg * res_w * res_w
    res     = tH_m1 * mp.exp( - arg2 / mp.mpf('8') ) / arg / mp.mpf('8')
    return res

def compute_tausb2(arg):
    res_w = func_w( tH )
    arg2    = arg * arg * res_w * res_w
    res     = tH_m2 * mp.exp( - arg2 / mp.mpf('8') ) / arg / mp.mpf('8')
    return res

def func_E( arg_s, arg_z ):
    arg = tz_R + arg_s * arg_z / mp.mpf('2') / tz_R
    res = tz_R * mp.erfc( arg )
    return res

def func_G( arg_s, arg_z ):
    arg = tz_R + arg_s * arg_z / mp.mpf('2') / tz_R
    res = mp.exp( - arg * arg ) / spi
    return res

def func_F( arg_s, arg_z ):
    res = mp.exp( arg_s * arg_z )
    return res

def func_G_F( arg_s, arg_z ):
    arg1 = tz_R + arg_s * arg_z / mp.mpf('2') / tz_R
    res = mp.exp( arg_s * arg_z - arg1 * arg1 ) / spi
    return res

def func_E_F( arg_s, arg_z ):
    arg1 = tz_R + arg_s * arg_z / mp.mpf('2') / tz_R
    res = mp.exp( arg_s * arg_z ) * ( tz_R * mp.erfc( arg1 ) )
    return res

def func_S( arg_s, arg_z ):
    arg1 = tz_R + arg_s * arg_z / mp.mpf('2') / tz_R
    arg2 = arg_s * arg_z 
    S1 = ( mp.mpf('1.5') - tz_R2 ) * tz_R2
    S2 = tz_R2 * ( arg1 * arg1 - mp.mpf('3') * arg2 / ( mp.mpf('4') * tz_R2 ) \
                  - mp.mpf('1') + mp.mpf('3') / ( mp.mpf('4') * tz_R2 ) )
    res_E = func_E( arg_s, arg_z )
    res_G = func_G( arg_s, arg_z )
    res = S1 * res_G + S2 * res_E
    return res

def func_V( arg_s, arg_z ):
   
    arg1 = tz_R + arg_s * arg_z / mp.mpf('2') / tz_R
    arg2 = arg_s * arg_z 
    V1 = tz_R2 * ( arg1 * arg1 - mp.mpf('0.5') * arg2 ) 
    V2 = tz_R2 * ( arg1 * arg1 - arg2 / ( mp.mpf('4') * tz_R2 ) ) 
    res_E = func_E( arg_s, arg_z )
    res_G = func_G( arg_s, arg_z )
    res = V1 * res_G - V2 * res_E
    res = res * ( - arg_s )
    return res

################## Script starts #########################
if show_version == 'y':
    print("numpy      version = ", np.__version__)
    print("matplotlib version = ", matplotlib.__version__)
    print("mpmath     version = ", mpmath.__version__)
    print("pandas     version = ", pd.__version__)

start = time.time() #

input_file, fname = get_fname(id)
read_parameters()
initialize()

for p in range(pmax):

    # In "search_s_max" mode, some parameters are replaced for fast computation.
    if search_s_max == 'y':
        if p == 0: # Index for the number of round in "search_s_max" mode
            start_search = time.time()
            err = mp.mpf('1e9')
            print('>> [COMMENT] search s_max start' )
            print('    now searching... err = '+"{:.2e}".format(float(err))+' current s_max = '+"{:.2e}".format(float(s_max)))
            # Remember the original parameters
            Nr_org = Nr; Nz_org = Nz; Nm_org = Nm; show_nd_org = show_nd; magnify_org = magnify; Nz_1_org = Nz_1; Nz_2_org = Nz_2
            # Set parameters for "search_s_max" mode
            Nr = Nr_ss; Nz = Nz_ss; Nm = Nm_ss; show_nd = 'n'; magnify = 'y'; Nz_1 = 1; Nz_2 = 1
            if Nm%2 != 1:
                Nm = Nm + 1
        else:
            if err > threshold_1_s_max:
                end_search = time.time()
                time_diff = end_search - start_search  
                print('    now searching... err = '+"{:.2e}".format(float(err))+' current s_max = '+"{:.2e}".format(float(s_max))+' ('+"{:.1f}".format(time_diff),'sec)')
                s_max = s_max * magnification_factor
            else:
                print('    search completed:err = '+"{:.2e}".format(float(err))+' current s_max = '+"{:.2e}".format(float(s_max)))
                print('\033[33m'+'>> [COMMENT] recommended "s_max0" in input file >= '+"{:.2e}".format(float(s_max/(z_R/H)))+'\033[0m')
                # Reload the original parameters
                Nr = Nr_org; Nz = Nz_org; Nm = Nm_org; show_nd = show_nd_org; magnify = magnify_org; Nz_1 = Nz_1_org;Nz_2 = Nz_2_org
                search_s_max = 'n'

    # In "search_Nm" mode, some parameters are replaced for fast computation. 
    if search_Nm == 'y' and search_s_max == 'n':
        if pp == -1: 
            pp = 0 # Index for the number of round in "search_Nm" mode
            start_search = time.time()
            err_Nm = mp.mpf('1e9')
            # Remember the original parameters
            Nr_org = Nr; Nz_org = Nz; Nm_org = Nm; show_nd_org = show_nd; magnify_org = magnify; Nz_1_org = Nz_1; Nz_2_org = Nz_2;
            # Set parameters for "search_Nm" mode
            Nr = Nr_sN; Nz = Nz_sN; Nm = Nm_sN; show_nd = 'n'; magnify = 'y'; Nz_1 = 1; Nz_2 = 1
            if Nm%2 != 1 and case == 1:
                Nm = Nm + 1
            Nm_array = []
            Nm_array = np.append(Nm_array,Nm)
            print('>> [COMMENT] search Nm start' )
        else:
            pp = pp + 1
            if err_Nm > threshold_1_Nm:
                end_search = time.time()
                time_diff = end_search - start_search  
                print('    now searching... err_Nm = '+"{:.2e}".format(float(err_Nm))+' current Nm = '+"{:3}".format(Nm)+' ('+"{:.1f}".format(time_diff),'sec)')
                if case == 1:
                    Nm = Nm - 1
                Nm = round( Nm * magnification_factor_Nm )
                if Nm%2 != 1 and case == 1:
                    Nm = Nm + 1
                show_nd = 'n'                
                Nm_array = np.append(Nm_array,Nm)
            else:
                print('    now searching... err_Nm = '+"{:.2e}".format(float(err_Nm))+' current Nm = '+"{:3}".format(Nm)+' ('+"{:.1f}".format(time_diff),'sec)')
                print('    search completed: max difference between Nm ='+"{:3}".format(Nm)+' and '+"{:3}".format(round(Nm_array[pp-3]))+' = '+"{:.2e}".format(float(err_Nm))  )
                if case == 1:
                    Nm = Nm - 1
                Nm = round(Nm/2)
                #print('    search completed:err_Nm = '+"{:.2e}".format(float(err_Nm))+' current Nm = '+"{:3}".format(Nm))
                print('\033[33m'+'>> [COMMENT] recommended "Nm" in input file >= '+"{:3}".format(Nm)+'\033[0m')
                # Reload the original parameters
                if Nm%2 != 1 and case == 1:
                    Nm = Nm + 1
                Nr = Nr_org; Nz = Nz_org; show_nd = show_nd_org; magnify = magnify_org; search_Nm = 'n'; Nz_1 = Nz_1_org; Nz_2 = Nz_2_org

    # Define lattice, compute macroscopic quantities, and show results.
    main()       

    # In not neither "search_s_max" nor "search_Nm" mode, exit. 
    if search_s_max != 'y' and search_Nm != 'y':
        break


end = time.time()
time_diff = end - start  
print('>> finish')
print('cpu-time = ',"{:.1f}".format(time_diff),'sec')

save_parameters()    


    


